import shlex

from IPython.core.magic import Magics, line_magic, magics_class, needs_local_scope


@magics_class
class HIMagics(Magics):

    @line_magic
    @needs_local_scope
    def objc(self, line, local_ns=None):
        """Load an Objective-C class by name into the IPython session."""
        p = local_ns.get('p')
        class_name = line.strip()
        if not class_name:
            p.log_error("Error: className is required.")
            return
        try:
            local_ns[class_name] = p.objc_get_class(class_name)
            p.log_info(f'{class_name} class loaded successfully')
        except Exception as e:
            p.log_error(f'Error loading class {class_name}: {str(e)}')

    @line_magic
    @needs_local_scope
    def fbp(self, line, local_ns=None):
        """Set a file breakpoint in the debugger."""
        p = local_ns.get('p')
        try:
            module_name, address = shlex.split(line.strip())
            address = int(address, 16)
            p.file_symbol(address, module_name).bp()
        except ValueError as ve:
            p.log_error(f"Error parsing arguments: {str(ve)}")
        except Exception as e:
            p.log_error(f"Error setting breakpoint: {str(e)}")


def load_ipython_extension(ipython):
    ipython.register_magics(HIMagics)



from queue import Queue

import pytest


@pytest.mark.parametrize('fmt, representation', [
    ('po', 'x2 = com.apple.powerlog.state_changed'),
    ('cf', 'x2 = <CFString'),
    ('x', 'x2 = 0x'),
    (lambda client, value: 'Just return this string', 'x2 = Just return this string')
])
def test_monitor_register_formats(hilda_client, fmt: str, representation: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param fmt: Register format.
    :param representation: Substring of register's representation.
    """
    if 'executable = aggregated' not in str(hilda_client.process):
        pytest.skip('Unsupported process, please run this test on aggregated')

    logs_queue = Queue()
    bp = hilda_client.symbols._PLStateChanged.monitor(regs={'x2': fmt}, stop=True)
    hilda_client.log_info = logs_queue.put_nowait
    # Trigger a call to _PLStateChanged.
    hilda_client.po('''
        @import ObjectiveC;
        @import Foundation;
        CFNotificationCenterRef center = CFNotificationCenterGetDarwinNotifyCenter();
        CFNotificationCenterPostNotification(center, CFSTR("com.apple.powerlog.state_changed"), NULL, NULL, FALSE);
    ''')
    hilda_client.cont()
    output = logs_queue.get(timeout=10)
    hilda_client.breakpoints.remove(bp)
    assert representation in output



from typing import Any, Callable, Generator, Optional, Union

import inquirer3

from hilda.lldb_importer import lldb
from hilda.symbol import Symbol

"""
A value identifying where the breakpoint was set (when it was created)

It could be either an address (int), a symbol name (string), a symbol name in a
specific module (Tuple[str, str], where the first item is the symbol name and the
second is the module name) or a Hilda symbol object (Symbol, that inherits from int).
"""
WhereType = Union[int, str, tuple[str, str], Symbol]


class HildaBreakpoint:
    """
    Hilda's class representing an LLDB breakpoint, with some optional additional properties
    """

    def __init__(self, hilda, lldb_breakpoint: lldb.SBBreakpoint,
                 where: Optional[WhereType] = None, description: Optional[str] = None) -> None:
        """
        Initialize a HildaBreakpoint.

        :param hilda.hilda_client.HildaClient hilda: Hilda client
        :param lldb.SBBreakpoint lldb_breakpoint: LLDB breakpoint to wrap
        :param WhereType where: Where the breakpoint is located
        :param description: Description of the breakpoint to appear upon `hilda.breakpoints.show()`
        """
        self._hilda = hilda
        self._where = where
        self._callback = None

        # Actual breakpoint from LLDB API
        self.lldb_breakpoint = lldb_breakpoint

        # If true, breakpoint will not be removed unless `remove_guarded` is requested
        self.guarded = False

        # Attach a description to the breakpoint
        self.description = description

    @property
    def where(self) -> Optional[WhereType]:
        """ Where the breakpoint was set (when it was created). """
        return self._where

    @property
    def id(self) -> int:
        """ A number identifying the breakpoint. """
        return self.lldb_breakpoint.GetID()

    @property
    def callback(self) -> Optional[Callable]:
        """
        A callback that will be executed when the breakpoint is hit.
        Note that unless the callback explicitly continues (by calling `cont()`), the program will not continue.
        The callback will be invoked as `callback(hilda, *args)`, where hilda is the `HildaClient`.
        """
        return self._callback

    @callback.setter
    def callback(self, callback: Optional[Callable]) -> None:
        self._callback = callback

        if callback is not None:
            self.lldb_breakpoint.SetScriptCallbackFunction(
                'lldb.hilda_client.breakpoints._dispatch_breakpoint_callback')

    @property
    def condition(self) -> Optional[str]:
        """ An LLDB expression to make this a conditional breakpoint. """
        return self.lldb_breakpoint.GetCondition()

    @condition.setter
    def condition(self, condition: Optional[str]) -> None:
        self.lldb_breakpoint.SetCondition(condition)

    @property
    def locations(self) -> list[lldb.SBBreakpointLocation]:
        """ LLDB locations array the breakpoint relates to. """
        return self.lldb_breakpoint.locations

    @property
    def name(self) -> Optional[str]:
        """
        A name for the breakpoint.
        When getting, if the breakpoint has multiple names, raises an exception.
        """
        names = self.names
        if len(names) == 0:
            return None
        if len(names) != 1:
            raise ValueError(f'Breakpoint {self.id} has multiple names {names}')

        name, = names
        return name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        self.names = {name}

    @property
    def names(self) -> set[str]:
        """ The set of names of the breakpoint. """
        name_list = lldb.SBStringList()
        self.lldb_breakpoint.GetNames(name_list)
        return set(name_list.GetStringAtIndex(i) for i in range(name_list.GetSize()))

    @names.setter
    def names(self, names: Union[set[str], list[str]]) -> None:
        new_names = set(names)
        if len(new_names) != len(names):
            raise ValueError(f'Duplicate names in {names}')

        current_names = self.names
        names_to_remove = current_names - new_names
        names_to_add = new_names - current_names
        for name in names_to_remove:
            self.lldb_breakpoint.RemoveName(name)
        for name in names_to_add:
            self.lldb_breakpoint.AddName(name)

    @property
    def enabled(self) -> bool:
        """
        Configures whether this breakpoint is enabled or not.
        """
        return self.lldb_breakpoint.IsEnabled()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.lldb_breakpoint.SetEnabled(value)

    def __repr__(self) -> str:
        enabled_repr = 'ENABLED' if self.enabled else 'DISABLED'
        guarded_repr = 'GUARDED' if self.guarded else 'NOT-GUARDED'
        return (f'<{self.__class__.__name__} LLDB:{self.lldb_breakpoint} {enabled_repr} {guarded_repr} '
                f'CALLBACK:{self.callback}>')

    def __str__(self) -> str:
        emoji = '🚨' if self.enabled else '🔕'
        enabled_str = 'enabled' if self.enabled else 'disabled'
        guarded_str = 'guarded' if self.guarded else 'not-guarded'

        result = f'{emoji} Breakpoint #{self.id} ({enabled_str}, {guarded_str})\n'

        if self.description is not None:
            result += f'\tDescription: {self.description}\n'

        if self.where is not None:
            result += f'\tWhere: {self.where}\n'

        # A single breakpoint may be related to several locations (addresses)
        locations = self.locations
        if len(locations) == 0:
            result += '\tNo locations\n'
        for location in self.locations:
            result += f'\tLocation {location}\n'

        return result.strip('\n')

    def remove(self, remove_guarded: bool = False) -> None:
        """
        Remove the breakpoint (unless the breakpoint is marked as guarded, see remove_guarded argument)

        :param bool remove_guarded: Remove the breakpoint even if the breakpoint is guarded
        """
        self._hilda.breakpoints.remove(self, remove_guarded)


class BreakpointList:
    """
    Manager for `HildaBreakpoint` objects, each one wrapping another native LLDB breakpoint.
    """

    def __init__(self, hilda) -> None:
        """
        Initialize a breakpoint list.

        :param hilda.hilda_client.HildaClient hilda: Hilda client
        """
        self._hilda = hilda
        self._breakpoints = {}

    def __contains__(self, id_or_name_or_bp: Union[int, str, HildaBreakpoint]):
        return self.get(id_or_name_or_bp) is not None

    def __iter__(self):
        for bp in self._hilda.target.breakpoint_iter():
            yield self[bp.GetID()]

    def __len__(self) -> int:
        return self._hilda.target.GetNumBreakpoints()

    def __getitem__(self, id_or_name_or_bp: Union[int, str, HildaBreakpoint]) -> HildaBreakpoint:
        """
        Get a breakpoint by ID or name (or the breakpoint itself, though it usually makes little sense)

        :param id_or_name_or_bp: Breakpoint's ID or name (or the breakpoint itself)
        """
        bp = self.get(id_or_name_or_bp)
        if bp is None:
            raise KeyError(id_or_name_or_bp)
        return bp

    def __delitem__(self, id_or_name_or_bp: Union[int, str, HildaBreakpoint]) -> None:
        """
        Remove a breakpoint (unless the breakpoint is marked as guarded - see remove())

        :param id_or_name_or_bp: Breakpoint's ID or name (or the breakpoint itself)
        """
        self.remove(id_or_name_or_bp)

    def __repr__(self) -> str:
        return repr(dict(self.items()))

    def __str__(self) -> str:
        return repr(self)

    def get(self, id_or_name_or_bp: Union[int, str, HildaBreakpoint]) -> Optional[HildaBreakpoint]:
        """
        Get a breakpoint by ID or name (or the breakpoint itself, though it usually makes little sense) or null
        if it does not exist.

        :param id_or_name_or_bp: Breakpoint's ID or name (or the breakpoint itself)
        :return: `HildaBreakpoint` if one exists, or `None` otherwise
        """

        if isinstance(id_or_name_or_bp, int):
            bp = self._hilda.target.FindBreakpointByID(id_or_name_or_bp)
        elif isinstance(id_or_name_or_bp, str):
            breakpoints = lldb.SBBreakpointList(self._hilda.target)
            found = self._hilda.target.FindBreakpointsByName(id_or_name_or_bp, breakpoints)
            if not found or breakpoints.GetSize() == 0:
                return None
            if breakpoints.GetSize() != 1:
                # Error out if we found multiple breakpoints with the name
                raise KeyError(id_or_name_or_bp)
            bp = breakpoints.GetBreakpointAtIndex(0)
        elif isinstance(id_or_name_or_bp, HildaBreakpoint):
            bp = id_or_name_or_bp.lldb_breakpoint
        else:
            raise TypeError()

        if not bp.IsValid():
            return None

        bp_id = bp.GetID()
        if bp_id not in self._breakpoints:
            self._hilda.log_debug(f'Found a breakpoint added outside of the Hilda API {bp}')
            self._breakpoints[bp_id] = HildaBreakpoint(self._hilda, bp)

        return self._breakpoints[bp_id]

    def add(self, where: WhereType, callback: Optional[Callable] = None, condition: str = None, guarded: bool = False,
            override: bool = True, description: Optional[str] = None) -> HildaBreakpoint:
        """
        Add a breakpoint.

        :param where: The address of the breakpoint.
            It could be either an address (int), a symbol name (string), a symbol name in a
            specific module (Tuple[str, str], where the first item is the symbol name and the
            second is the module name) or a Hilda symbol object (Symbol, that inherits from int).
        :param callback: A callback that will be executed when the breakpoint is hit.
            Note that unless the callback explicitly continues (by calling cont()), the program will not continue.
            The callback will be invoked as callback(hilda, *args), where hilda is the 'HildaClient'.
        :param condition: An LLDB expression to make this a conditional breakpoint.
        :param guarded: If true, breakpoint will not be removed unless remove_guarded is requested
        :param override: If True and an existing breakpoint with the same `where` is found, remove the old
            breakpoint and replace it with the new breakpoint. Otherwise, prompt the user.
        :return: The new breakpoint
        """
        if where in (bp.where for bp in self):
            if override or inquirer3.confirm('A breakpoint already exist in given location. '
                                             'Would you like to delete the previous one?', True):
                for bp in list(bp for bp in self if where == bp.where):
                    del self[bp]

        if isinstance(where, int):
            bp = self._hilda.target.BreakpointCreateByAddress(where)
        elif isinstance(where, str):
            # Note that the name in BreakpointCreateByName is the name of the location,
            # not the name of the breakpoint.
            bp = self._hilda.target.BreakpointCreateByName(where)
        elif isinstance(where, tuple):
            name, module = where
            raise NotImplementedError()
        if not bp.IsValid():
            raise Exception(f'Failed to create breakpoint at {where}')

        bp = HildaBreakpoint(self._hilda, bp, where, description=description)
        bp.callback = callback
        bp.condition = condition
        bp.guarded = guarded

        self._breakpoints[bp.id] = bp

        self._hilda.log_info(f'Breakpoint #{bp.id} has been set')
        return bp

    def add_monitor(self, where: WhereType,
                    condition: str = None,
                    guarded: bool = False,
                    override: bool = True,
                    regs: Optional[dict[str, Union[str, Callable]]] = None,
                    expr: Optional[dict[str, Union[str, Callable]]] = None,
                    retval: Optional[Union[str, Callable]] = None,
                    stop: bool = False,
                    bt: bool = False,
                    cmd: Optional[list[str]] = None,
                    force_return: Optional[bool] = None,
                    name: Optional[str] = None,
                    description: Optional[str] = None,
                    ) -> HildaBreakpoint:
        """
        Monitor every time a given address is called.

        Creates a breakpoint whose callback implements the requested features.

        :param where: See add() for details.
        :param condition: See add() for details.
        :param guarded: See add() for details.
        :param override: See add() for details.
        :param regs: Print register values (using the provided format).
            E.g., `regs={'x0': 'x'}` prints x0 in HEX format
            The available formats are:
                'x': hex
                's': string
                'cf': use CFCopyDescription() to get more informative description of the object
                'po': use LLDB po command
                'std::string': for std::string
                Callable: user defined function, will be called like `format_function(hilda_client, value)`.
        :param expr: Print LLDB expression (using the provided format).
            E.g., `expr={'$x0': 'x', '$arg1': 'x'}` (to print the value of x0 and arg1).
            The format behaves like in regs above.
        :param retval: Print the return value of the function (using the provided format).
            The format behaves like in regs above.
        :param stop: If True, stop whenever the breakpoint is hit (otherwise continue debugging).
        :param bt: Print backtrace.
        :param cmd: A list of LLDB commands to run when the breakpoint is hit.
        :param force_return: Return immediately from the function, returning the specified value.
        :param name: Use the provided name instead of the symbol name automatically extracted from the calling frame.
        :param description: Attach a brief description of the breakpoint.
        :return: The new breakpoint
        """

        if regs is None:
            regs = {}
        if expr is None:
            expr = {}
        if cmd is None:
            cmd = []

        def callback(hilda, frame: lldb.SBFrame, bp_loc: lldb.SBBreakpointLocation, *_) -> None:
            """
            Callback function called when a breakpoint is hit.

            :param hilda.hilda_client.HildaClient hilda: Hilda client to operate on
            :param frame: LLDB frame
            :param bp_loc: LLDB breakpoint location
            """
            nonlocal name
            bp = bp_loc.GetBreakpoint()
            symbol = hilda.symbol(hilda.frame.addr.GetLoadAddress(hilda.target))
            thread = hilda.thread
            printed_name = name if name is not None else str(symbol.lldb_address)

            def format_value(fmt: Union[str, Callable], value: Symbol) -> str:
                if callable(fmt):
                    return fmt(hilda, value)
                formatters = {
                    'x': lambda val: f'0x{int(val):x}',
                    's': lambda val: val.peek_str() if val else None,
                    'cf': lambda val: val.cf_description,
                    'po': lambda val: val.po(),
                    'std::string': hilda._std_string
                }
                if fmt in formatters:
                    return formatters[fmt](value)
                else:
                    return f'{value:x} (unsupported format)'

            log_message = f'🚨 #{bp.id} 0x{symbol:x} {printed_name} - Thread #{thread.idx}:{hex(thread.id)}'

            if regs != {}:
                log_message += '\nregs:'
                for name, fmt in regs.items():
                    value = hilda.symbol(frame.FindRegister(name).unsigned)
                    log_message += f'\n\t{name} = {format_value(fmt, value)}'

            if expr != {}:
                log_message += '\nexpr:'
                for name, fmt in expr.items():
                    value = hilda.symbol(hilda.evaluate_expression(name))
                    log_message += f'\n\t{name} = {format_value(fmt, value)}'

            if force_return is not None:
                hilda.force_return(force_return)
                log_message += f'\nforced return: {force_return}'

            if bt:
                # bugfix: for callstacks from xpc events
                hilda.finish()
                for frame in hilda.bt():
                    log_message += f'\n\t{frame[0]} {frame[1]}'

            if retval is not None:
                # return from function
                hilda.finish()
                value = hilda.evaluate_expression('$arg1')
                log_message += f'\nreturned: {format_value(retval, value)}'

            hilda.log_info(log_message)

            for command in cmd:
                hilda.lldb_handle_command(command)

            if stop:
                hilda.log_info('Process remains stopped and focused on current thread')
            else:
                hilda.cont()

        return self.add(where, callback, condition=condition, guarded=guarded, override=override,
                        description=description)

    def remove(self, id_or_name_or_bp: Union[int, str, HildaBreakpoint], remove_guarded: bool = False) -> None:
        """
        Remove a breakpoint (unless the breakpoint is marked as guarded, see remove_guarded argument).

        :param id_or_name_or_bp: Breakpoint's ID or name (or the breakpoint itself)
        :param remove_guarded: Remove breakpoint even if the breakpoint is marked as guarded
        """

        bp = self[id_or_name_or_bp]

        if bp.guarded and not remove_guarded:
            self._hilda.log_warning(f'Remove request for breakpoint {bp} is ignored')
            return

        # Removing a breakpoint without using this function would leak the breakpoint in self._breakpoints

        breakpoint_id = bp.id
        self._hilda.target.BreakpointDelete(breakpoint_id)
        self._hilda.log_debug(f'Breakpoint #{breakpoint_id} has been removed')

    def clear(self, remove_guarded: bool = False) -> None:
        """
        Remove all breakpoints (except for breakpoints marked as guarded, see remove_guarded argument).

        :param remove_guarded: Also remove breakpoints marked as guarded
        """
        breakpoints = list(self)
        for bp in breakpoints:
            if not remove_guarded and bp.guarded:
                continue

            self.remove(bp, remove_guarded)

    def show(self) -> None:
        """ Show existing breakpoints. """
        if len(self) == 0:
            self._hilda.log_info('No breakpoints')
        for bp in self:
            self._hilda.log_info(bp)

    def items(self):
        """
        Get a breakpoint ID and breakpoint object tuple for every breakpoint
        """
        return ((bp.id, bp) for bp in self)

    def keys(self) -> Generator[int, Any, None]:
        """
        Get the breakpoint ID for every breakpoint
        """
        return (bp.id for bp in self)

    def values(self) -> Generator[HildaBreakpoint, Any, None]:
        """
        Get the breakpoint object for every breakpoint
        """
        return (bp for bp in self)

    def _dispatch_breakpoint_callback(self, frame, bp_loc, *_) -> None:
        """
        Route the breakpoint callback the specific breakpoint callback.

        :param lldb.SBFrame frame: LLDB Frame object.
        :param lldb.SBBreakpointLocation bp_loc: LLDB Breakpoint location object.
        """

        bp_id = bp_loc.GetBreakpoint().GetID()
        self._hilda._bp_frame = frame
        try:
            callback = self[bp_id].callback
            if callback is not None:
                callback(self._hilda, frame, bp_loc, self[bp_id])
        finally:
            self._hilda._bp_frame = None



import pytest

from hilda.hilda_client import HildaClient
from hilda.launch_lldb import create_hilda_client_using_attach_by_name

PROCESS = 'remoted'


@pytest.fixture(scope='function')
def hilda_client() -> HildaClient:
    with create_hilda_client_using_attach_by_name(PROCESS) as hilda_client:
        yield hilda_client



import pytest

from hilda.exceptions import SymbolAbsentError
from hilda.symbols import SymbolList


def test_filter_by_module(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    assert 'isdigit' in libsystem_c_symbols

    libobjc_symbols = hilda_client.symbols.filter_by_module('libobjc')
    assert 'isdigit' not in libobjc_symbols

    assert 'isdigit' in hilda_client.symbols


def test_iter(hilda_client):
    # Test that iteration does not fail
    i = 0
    for symbol in hilda_client.symbols:
        i += 1


def test_getitem(hilda_client):
    isdigit = hilda_client.symbols['isdigit']
    assert not isdigit(0x10)
    assert isdigit(0x30)


def test_delitem(hilda_client):
    symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    assert 'isdigit' in symbols

    isdigit = symbols.isdigit
    del symbols[isdigit]

    assert 'isdigit' not in symbols

    try:
        del symbols[isdigit]
        assert False
    except:  # noqa: E722
        pass

    assert 'isdigit' in hilda_client.symbols


def test_get(hilda_client):
    isdigit = hilda_client.symbols.get('isdigit')
    assert isdigit is not None

    assert hilda_client.symbols.get(int(isdigit)) is isdigit
    assert hilda_client.symbols.get('isdigit') is isdigit
    assert hilda_client.symbols.get(isdigit.id) is isdigit
    assert hilda_client.symbols.get(isdigit.lldb_symbol) is isdigit
    assert hilda_client.symbols.get(isdigit) is isdigit
    assert hilda_client.symbols.isdigit is isdigit

    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    assert libsystem_c_symbols.get(isdigit) is isdigit

    assert hilda_client.symbols.get(isdigit + 4) is None
    assert libsystem_c_symbols.get(isdigit + 4) is None


def test_get_absent_symbol(hilda_client):
    with pytest.raises(SymbolAbsentError):
        hilda_client.symbols.symbol_that_doesnt_exist


def test_add_symbol(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    isdigit = libsystem_c_symbols.isdigit

    assert libsystem_c_symbols.get(isdigit + 4) is None
    libsystem_c_symbols.add(isdigit + 4, 'test_symbol_in_isdigit_added_to_libsystem_c')
    assert libsystem_c_symbols.get(isdigit + 4) is not None
    assert libsystem_c_symbols.get('test_symbol_in_isdigit_added_to_libsystem_c') is not None

    assert hilda_client.symbols.get(isdigit + 4) is not None
    assert hilda_client.symbols.get('test_symbol_in_isdigit_added_to_libsystem_c') is not None


def test_getattr(hilda_client):
    assert hilda_client.symbols.x11223344 is not None
    assert hilda_client.symbols.x0x11223344 is not None
    assert hilda_client.symbols.malloc is not None


def test_sub_symbol_lists(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    l1 = SymbolList(hilda_client)
    l1.add(libsystem_c_symbols.isdigit)
    l2 = libsystem_c_symbols - l1
    assert libsystem_c_symbols.isdigit not in l2


def test_add_symbol_lists(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')

    l1 = SymbolList(hilda_client)
    l1.add(libsystem_c_symbols.isdigit)
    l2 = SymbolList(hilda_client)
    l3 = l2 + l1
    assert libsystem_c_symbols.isdigit in l3


def test_filter_symbol_type(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    assert libsystem_c_symbols.isdigit in libsystem_c_symbols.filter_code_symbols()
    assert libsystem_c_symbols.isdigit not in libsystem_c_symbols.filter_data_symbols()


def test_filter_startswith(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    filtered = libsystem_c_symbols.filter_startswith('is')
    assert libsystem_c_symbols.isdigit in filtered
    assert libsystem_c_symbols.isalpha in filtered
    assert libsystem_c_symbols.rand not in filtered


def test_filter_endswith(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    filtered = libsystem_c_symbols.filter_endswith('digit')
    assert libsystem_c_symbols.isdigit in filtered
    assert libsystem_c_symbols.isalpha not in filtered
    assert libsystem_c_symbols.rand not in filtered


def test_filter_name_contains(hilda_client):
    libsystem_c_symbols = hilda_client.symbols.filter_by_module('libsystem_c')
    filtered = libsystem_c_symbols.filter_name_contains('dig')
    assert libsystem_c_symbols.isdigit in filtered
    assert libsystem_c_symbols.isalpha not in filtered
    assert libsystem_c_symbols.rand not in filtered
    assert libsystem_c_symbols.isdigit in libsystem_c_symbols.filter_name_contains('is')



from dataclasses import dataclass
from typing import Callable

from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import EmacsInsertMode, HasFocus, HasSelection, ViInsertMode
from prompt_toolkit.keys import Keys


@dataclass
class Keybinding:
    key: str
    description: str
    callback: Callable


def get_keybindings(hilda_client) -> list[Keybinding]:
    """
    Get list of keybindings

    :param hilda.hilda_client.HildaClient hilda_client: Hilda client to bind the keys to operations
    """
    return [
        Keybinding(key=Keys.F1, description='Show this help', callback=hilda_client.show_help),
        Keybinding(key=Keys.F2, description='Show process state UI', callback=hilda_client.ui_manager.show),
        Keybinding(key=Keys.F3, description='Toggle enabling of stdout & stderr',
                   callback=hilda_client.toggle_enable_stdout_stderr),
        Keybinding(key=Keys.F7, description='Step Into', callback=hilda_client.step_into),
        Keybinding(key=Keys.F8, description='Step Over', callback=hilda_client.step_over),
        Keybinding(key=Keys.F9, description='Continue',
                   callback=lambda _: (hilda_client.log_info('Sending continue'), hilda_client.cont())),
        Keybinding(key=Keys.F10, description='Stop',
                   callback=lambda _: (hilda_client.log_info('Sending stop'), hilda_client.stop())),
    ]


def load_ipython_extension(ipython):
    def register_keybindings():
        hilda_client = ipython.user_ns['p']
        insert_mode = ViInsertMode() | EmacsInsertMode()
        registry = ipython.pt_app.key_bindings

        for keybind in get_keybindings(hilda_client):
            registry.add_binding(
                keybind.key, filter=(HasFocus(DEFAULT_BUFFER) & ~HasSelection() & insert_mode))(keybind.callback)

    register_keybindings()
    ipython.events.register('shell_initialized', register_keybindings)



import base64
import builtins
import importlib
import importlib.util
import json
import logging
import os
import struct
import sys
import time
import typing
from collections import namedtuple
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import cached_property, wraps
from pathlib import Path
from typing import Any, Callable, Optional, Union

import click
import hexdump
import IPython
from IPython.core.magic import register_line_magic  # noqa: F401
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import XmlLexer
from tqdm import tqdm
from traitlets.config import Config

from hilda import objective_c_class
from hilda.breakpoints import BreakpointList, HildaBreakpoint, WhereType
from hilda.common import CfSerializable, selection_prompt
from hilda.exceptions import AccessingMemoryError, AccessingRegisterError, ConvertingFromNSObjectError, \
    ConvertingToNsObjectError, CreatingObjectiveCSymbolError, DisableJetsamMemoryChecksError, \
    EvaluatingExpressionError, HildaException, InvalidThreadIndexError, SymbolAbsentError
from hilda.ipython_extensions.keybindings import get_keybindings
from hilda.lldb_importer import lldb
from hilda.objective_c_symbol import ObjectiveCSymbol
from hilda.registers import Registers
from hilda.snippets.mach import CFRunLoopServiceMachPort_hooks
from hilda.symbol import Symbol
from hilda.symbols import SymbolList
from hilda.ui.ui_manager import UiManager
from hilda.watchpoints import WatchpointList

lldb.KEYSTONE_SUPPORT = True
try:
    from keystone import KS_ARCH_ARM64, KS_ARCH_X86, KS_MODE_64, KS_MODE_LITTLE_ENDIAN, Ks
except ImportError:
    lldb.KEYSTONE_SUPPORT = False
    print('failed to import keystone. disabling some features')


def disable_logs() -> None:
    logging.getLogger('asyncio').disabled = True
    logging.getLogger('parso.cache').disabled = True
    logging.getLogger('parso.cache.pickle').disabled = True
    logging.getLogger('parso.python.diff').disabled = True
    logging.getLogger('blib2to3.pgen2.driver').disabled = True
    logging.getLogger('hilda.launch_lldb').setLevel(logging.INFO)


SerializableSymbol = namedtuple('SerializableSymbol', 'address type_ filename')


@dataclass
class HelpSnippet:
    """ Small snippet line to occur from `HildaClient.show_help()` """

    key: str
    description: str

    def __str__(self) -> str:
        return click.style(self.key.ljust(8), bold=True, fg='magenta') + click.style(self.description, bold=True)


@dataclass
class Configs:
    """ Configuration settings for evaluation and monitoring. """
    evaluation_unwind_on_error: bool = field(default=False,
                                             metadata={'doc': 'Whether to unwind on error during evaluation.'})
    evaluation_ignore_breakpoints: bool = field(default=False,
                                                metadata={'doc': 'Whether to ignore breakpoints during evaluation.'})
    nsobject_exclusion: bool = field(default=False, metadata={
        'doc': 'Whether to exclude NSObject during evaluation - reduce ipython autocomplete results.'})
    objc_verbose_monitor: bool = field(default=False, metadata={
        'doc': 'When set to True, using monitor() will automatically print objc methods arguments.'})
    enable_stdout_stderr: bool = field(default=True, metadata={
        'doc': 'When set to True, will enable process stdout and stderr.'})

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        config_str = 'Configuration settings:\n'
        max_len = max(len(field_name) for field_name in self.__dataclass_fields__) + 2

        for field_name, field_info in self.__dataclass_fields__.items():
            value = getattr(self, field_name)
            doc = field_info.metadata.get('doc', 'No docstring available')
            config_str += f'\t{field_name.ljust(max_len)}: {str(value).ljust(5)} | {doc}\n'

        return config_str


def stop_is_needed(func: Callable):
    """Decorator to check if the process must be stopped before proceeding."""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        is_running = self.process.GetState() == lldb.eStateRunning
        if is_running:
            self.logger.error(f'Cannot {func.__name__.replace("_", "-")}: Process must be stopped first.')
            return
        return func(self, *args, **kwargs)

    return wrapper


class HildaClient:
    RETVAL_BIT_COUNT = 64

    def __init__(self, debugger: lldb.SBDebugger) -> None:
        self.logger = logging.getLogger(__name__)
        self.endianness = '<'
        self.debugger = debugger
        self.target = debugger.GetSelectedTarget()
        self.process = self.target.GetProcess()
        self.symbols = SymbolList(self)
        self.breakpoints = BreakpointList(self)
        self.watchpoints = WatchpointList(self)
        self.captured_objects = {}
        self.registers = Registers(self)
        self.arch = self.target.GetTriple().split('-')[0]
        self.ui_manager = UiManager(self)
        self.configs = Configs()
        self._dynamic_env_loaded = False
        self._symbols_loaded = False
        self.globals: typing.MutableMapping[str, Any] = globals()
        self._hilda_root = Path(__file__).parent

        # the frame called within the context of the hit BP
        self._bp_frame = None

        self._add_global('symbols', self.symbols, [])
        self._add_global('registers', self.registers, [])

        self.log_info(f'Target: {self.target}')
        self.log_info(f'Process: {self.process}')

    def hd(self, buf: bytes) -> None:
        """
        Print hexdump representation for given buffer.

        :param buf: buffer to print in hexdump form
        """
        hexdump.hexdump(buf)

    def lsof(self) -> dict[int, Any]:
        """
        Get dictionary of all open FDs
        :return: Mapping between open FDs and their paths
        """
        data = (self._hilda_root / 'objective_c' / 'lsof.m').read_text()
        result = json.loads(self.po(data))
        # convert FDs into int
        return {int(k): v for k, v in result.items()}

    def bt(self, should_print: bool = False, depth: Optional[int] = None) -> list[Union[str, lldb.SBFrame]]:
        """ Print an improved backtrace. """
        backtrace = []
        for i, frame in enumerate(self.thread.frames):
            if i == depth:
                break
            row = ''
            row += click.style(f'0x{frame.addr.GetFileAddress():x} ', fg='cyan')
            row += str(frame)
            if i == 0:
                # first line
                row += ' 👈'
            backtrace.append([f'0x{frame.addr.file_addr:016x}', frame])
            if should_print:
                print(row)
        return backtrace

    def disable_jetsam_memory_checks(self):
        """
        Disable jetsam memory checks, prevent raising:
        `error: Execution was interrupted, reason: EXC_RESOURCE RESOURCE_TYPE_MEMORY (limit=15 MB, unused=0x0).`
        when evaluating expression.
        """
        # 6 is for MEMORYSTATUS_CMD_SET_JETSAM_TASK_LIMIT
        result = self.symbols.memorystatus_control(6, self.process.GetProcessID(), 0, 0, 0)
        if result:
            raise DisableJetsamMemoryChecksError()

    def symbol(self, address: int) -> Symbol:
        """
        Get symbol object for a given address
        :param address:
        :return: Hilda's symbol object
        """
        return self.symbols.add(address)

    def objc_symbol(self, address: int) -> ObjectiveCSymbol:
        """
        Get objc symbol wrapper for given address
        :param address:
        :return: Hilda's objc symbol object
        """
        try:
            return ObjectiveCSymbol.create(int(address), self)
        except HildaException as e:
            raise CreatingObjectiveCSymbolError from e

    def inject(self, filename: str) -> SymbolList:
        """
        Inject a single library into currently running process.

        :param filename: library to inject (dylib)
        :return: SymbolList
        """
        module = self.target.FindModule(lldb.SBFileSpec(os.path.basename(filename), False))
        if module.file.basename is not None:
            self.log_warning(f'file {filename} has already been loaded')

        injected = SymbolList(self)
        handle = self.symbols.dlopen(filename, 10)  # RTLD_GLOBAL|RTLD_NOW

        if handle == 0:
            self.log_critical(f'failed to inject: {filename}')

        module = self.target.FindModule(lldb.SBFileSpec(os.path.basename(filename), False))
        for symbol in module.symbols:
            load_addr = symbol.addr.GetLoadAddress(self.target)
            if load_addr == 0xffffffffffffffff:
                # skip those not having a real address
                continue

            name = symbol.name
            type_ = symbol.GetType()

            if name in ('<redacted>',) or (type_ not in (lldb.eSymbolTypeCode,
                                                         lldb.eSymbolTypeData,
                                                         lldb.eSymbolTypeObjCMetaClass)):
                # ignore unnamed symbols and those which are not: data, code or objc classes
                continue

            injected.add(self.symbol(load_addr), name)
        return injected

    @stop_is_needed
    def poke(self, address, buf: bytes):
        """
        Write data at given address
        :param address:
        :param buf:
        """
        err = lldb.SBError()
        retval = self.process.WriteMemory(address, buf, err)

        if not err.Success():
            raise AccessingMemoryError()

        return retval

    @stop_is_needed
    def poke_text(self, address: int, code: str) -> int:
        """
        Write instructions to address.
        :param address:
        :param code:
        """
        if not lldb.KEYSTONE_SUPPORT:
            raise NotImplementedError('Not supported without keystone')
        bytecode, count = self._ks.asm(code, as_bytes=True)
        return self.poke(address, bytecode)

    @stop_is_needed
    def peek(self, address, size: int) -> bytes:
        """
        Read data at given address
        :param address:
        :param size:
        :return:
        """
        if size == 0:
            return b''

        err = lldb.SBError()
        retval = self.process.ReadMemory(address, int(size), err)

        if not err.Success():
            raise AccessingMemoryError()

        return retval

    @stop_is_needed
    def peek_str(self, address: Symbol) -> str:
        """
        Peek a buffer till null termination
        :param address:
        :return:
        """
        return address.po('char *')[1:-1]  # strip the ""

    @stop_is_needed
    def peek_std_str(self, address: Symbol) -> str:
        """
        Peek a std::string
        :param address: Address to read the std::string from
        :return: Python string
        """
        return self._std_string(address)

    def stop(self, *args) -> None:
        """ Stop process. """
        self.debugger.SetAsync(False)

        is_running = self.process.GetState() == lldb.eStateRunning
        if not is_running:
            self.log_debug('already stopped')
            return

        if not self.process.Stop().Success():
            self.log_critical('failed to stop process')

    def run_for(self, seconds: float) -> None:
        """
        Run the process for a given time
        :return:
        """
        self.cont()
        self.logger.info(f'Running for {seconds} seconds')
        time.sleep(seconds)
        self.stop()

    def cont(self, *args) -> None:
        """ Continue process. """
        is_running = self.process.GetState() == lldb.eStateRunning

        if is_running:
            self.log_debug('already running')
            return

        # bugfix:   the debugger may become in sync state, so we make sure
        #           it isn't before trying to continue
        self.debugger.SetAsync(True)

        if not self.process.Continue().Success():
            self.log_critical('failed to continue process')

    def detach(self) -> None:
        """
        Detach from process.

        Useful in order to exit gracefully so process doesn't get killed
        while you exit
        """
        if not self.process.is_alive:
            return
        if not self.process.Detach().Success():
            self.log_critical('failed to detach')
            return
        self.log_info('Process Detached')

    @stop_is_needed
    def disass(self, address: int, buf: bytes, flavor: str = 'intel',
               should_print: bool = False) -> lldb.SBInstructionList:
        """
        Print disassembly from a given address
        :param flavor:
        :param address:
        :param buf:
        :param should_print:
        :return:
        """
        inst = self.target.GetInstructionsWithFlavor(lldb.SBAddress(address, self.target), flavor, buf)
        if should_print:
            print(inst)
        return inst

    def file_symbol(self, address: int, module_name: Optional[str] = None) -> Symbol:
        """
        Calculate symbol address without ASLR
        :param address: address as can be seen originally in Mach-O
        :param module_name: Module name to resolve the symbol from
        """
        if module_name is None:
            module = self.target
        else:
            module = self.target.FindModule(lldb.SBFileSpec(module_name))

        return self.symbol(module.ResolveFileAddress(address).GetLoadAddress(self.target))

    def get_register(self, name: str) -> Union[float, Symbol]:
        """
        Get value for register by its name. Value can either be an Symbol (int) or a float.

        :param name: Register name
        :return: Register value
        """
        register_value = self.frame.register[name.lower()]
        if register_value is None:
            raise AccessingRegisterError()
        return self._get_symbol_or_float_from_sbvalue(register_value)

    def set_register(self, name: str, value: Union[float, int]) -> None:
        """
        Set value for register by its name.

        :param name: Register name
        :param value: Register value
        """
        register = self.frame.register[name.lower()]
        if register is None:
            raise AccessingRegisterError()
        if isinstance(value, int):
            register.value = hex(value)
        else:
            register.value = str(value)

    def objc_call(self, obj: int, selector: str, *params):
        """
        Simulate a call to an objc selector.

        :param obj: obj to pass into `objc_msgSend`
        :param selector: selector to execute
        :param params: any other additional parameters the selector requires
        :return: invocation returned value
        """
        # On object `obj`
        args = self._serialize_call_params([obj])
        # Call selector (by its uid)
        args.append(self._generate_call_expression(self.symbols.sel_getUid, self._serialize_call_params([selector])))
        # With params
        args.extend(self._serialize_call_params(params))
        call_expression = self._generate_call_expression(self.symbols.objc_msgSend, args)
        with self.stopped():
            return self.evaluate_expression(call_expression)

    def call(self, address, argv: Optional[list] = None):
        """
        Call function at given address with given parameters
        :param address:
        :param argv: parameter list
        :return: function's return value
        """
        if argv is None:
            argv = []
        call_expression = self._generate_call_expression(address, self._serialize_call_params(argv))
        with self.stopped():
            return self.evaluate_expression(call_expression)

    def monitor(self, address, condition: str = None, **options) -> HildaBreakpoint:
        """
        Monitor every time a given address is called

        Alias of self.breakpoints.add_monitor()
        """
        return self.breakpoints.add_monitor(address, condition, **options)

    def show_current_source(self) -> None:
        """ print current source code if possible """
        self.lldb_handle_command('f')

    def finish(self):
        """ Run current frame till its end. """
        with self.sync_mode():
            self.thread.StepOutOfFrame(self.frame)
            self._bp_frame = None

    @stop_is_needed
    def step_into(self, *args):
        """ Step into current instruction. """
        with self.sync_mode():
            self.thread.StepInto()
        if self.ui_manager.active:
            self.ui_manager.show()

    @stop_is_needed
    def step_over(self, *args):
        """ Step over current instruction. """
        with self.sync_mode():
            self.thread.StepOver()
        if self.ui_manager.active:
            self.ui_manager.show()

    def force_return(self, value: int = 0) -> None:
        """
        Prematurely return from a stack frame, short-circuiting exection of newer frames and optionally
        yielding a specified value.
        :param value:
        :return:
        """
        self.finish()
        self.set_register('x0', value)

    def proc_info(self) -> None:
        """ Print information about currently running mapped process. """
        print(self.process)

    def print_proc_entitlements(self) -> None:
        """ Get the plist embedded inside the process' __LINKEDIT section. """
        linkedit_section = self.target.modules[0].FindSection('__LINKEDIT')
        linkedit_data = self.symbol(linkedit_section.GetLoadAddress(self.target)).peek(linkedit_section.size)

        # just look for the xml start inside the __LINKEDIT section. should be good enough since wer'e not
        # expecting any other XML there
        entitlements = str(linkedit_data[linkedit_data.find(b'<?xml'):].split(b'\xfa', 1)[0], 'utf8')
        print(highlight(entitlements, XmlLexer(), TerminalTrueColorFormatter()))

    def bp(self, address_or_name: WhereType, callback: Optional[Callable] = None, condition: Optional[str] = None,
           guarded: bool = False, description: Optional[str] = None, **options) -> HildaBreakpoint:
        """
        Add a breakpoint

        Alias of self.breakpoints.add()

        :param address_or_name: Where to place the breakpoint
        :param condition: set as a conditional breakpoint using lldb expression
        :param callback: callback(hilda, *args) to be called
        :param guarded: whether the breakpoint should be protected frm usual removal.
        :param description: Attach a breakpoint description
        :param options: can contain an `override` keyword to specify if to override an existing BP
        :return: native LLDB breakpoint
        """
        return self.breakpoints.add(address_or_name, callback, condition, guarded, description=description, **options)

    def po(self, expression: str, cast: Optional[str] = None) -> str:
        """
        Print given object using LLDB's po command

        Can also run big chunks of native code:

        po('NSMutableString *s = [NSMutableString string]; [s appendString:@"abc"]; [s description]')

        :param expression: either a symbol or string the execute
        :param cast: object type
        :raise EvaluatingExpressionError: LLDB failed to evaluate the expression
        :return: LLDB's po output
        """
        casted_expression = ''
        if cast is not None:
            casted_expression += '(%s)' % cast
        casted_expression += f'0x{expression:x}' if isinstance(expression, int) else str(expression)

        res = lldb.SBCommandReturnObject()
        self.debugger.GetCommandInterpreter().HandleCommand(f'expression -i 0 -lobjc -O -- {casted_expression}', res)
        if not res.Succeeded():
            raise EvaluatingExpressionError(res.GetError())
        return res.GetOutput().strip()

    def globalize_symbols(self) -> None:
        """
        Make all symbols in python's global scope
        """
        reserved_names = list(globals().keys()) + dir(builtins)
        for name, value in tqdm(self.symbols.items()):
            if ':' not in name \
                    and '[' not in name \
                    and '<' not in name \
                    and '(' not in name \
                    and '.' not in name:
                self._add_global(name, value, reserved_names)

    def jump(self, symbol: int) -> None:
        """ jump to given symbol """
        self.lldb_handle_command(f'j *{symbol}')

    def lldb_handle_command(self, cmd: str, capture_output: bool = False) -> Optional[str]:
        """
        Execute an LLDB command

        For example:
            lldb_handle_command('register read')

        :param cmd: LLDB command
        :param capture_output: True if capturing the command output
        :return: The output if capture was requested, None if not or the command failed
        """
        if capture_output:
            result = lldb.SBCommandReturnObject()
            self.debugger.GetCommandInterpreter().HandleCommand(cmd, result)
            return result.GetOutput() if result.Succeeded() else None
        else:
            self.debugger.HandleCommand(cmd)

    def objc_get_class(self, name: str, module_name: Optional[str] = None) -> objective_c_class.Class:
        """
        Get ObjC class object
        :param module_name:
        :param name:
        :return:
        """
        if module_name is not None:
            ret = self.symbol(self._get_module_class_list(module_name)[name]).objc_class
        else:
            ret = objective_c_class.Class.from_class_name(self, name)
        return ret

    def CFSTR(self, symbol: int) -> Symbol:
        """ Create CFStringRef object from given string """
        return self.cf(symbol)

    def cf(self, data: CfSerializable) -> Symbol:
        """
        Create NSObject from given data (same as ns())
        :param data: Data representing the NSObject, must by JSON serializable
        :return: Pointer to a NSObject
        """
        return self.ns(data)

    def ns(self, data: CfSerializable) -> Symbol:
        """
        Create NSObject from given data (same as cf())
        :param data: Data representing the NSObject, must by JSON serializable
        :return: Pointer to a NSObject
        """
        try:
            json_data = json.dumps({'root': data}, default=self._to_ns_json_default)
        except TypeError as e:
            raise ConvertingToNsObjectError from e
        obj_c_code = (self._hilda_root / 'objective_c' / 'to_ns_from_json.m').read_text()
        expression = obj_c_code.replace('__json_object_dump__', json_data.replace('"', r'\"'))
        try:
            return self.evaluate_expression(expression)
        except EvaluatingExpressionError as e:
            raise ConvertingToNsObjectError from e

    def decode_cf(self, address: Union[int, str]) -> CfSerializable:
        """
        Create python object from NS object.
        :param address: NS object.
        :return: Python object.
        """
        obj_c_code = (self._hilda_root / 'objective_c' / 'from_ns_to_json.m').read_text()
        address = f'0x{address:x}' if isinstance(address, int) else address
        expression = obj_c_code.replace('__ns_object_address__', address)
        try:
            json_dump = self.po(expression)
        except EvaluatingExpressionError as e:
            raise ConvertingFromNSObjectError from e
        return json.loads(json_dump, object_hook=self._from_ns_json_object_hook)['root']

    def evaluate_expression(self, expression: str) -> Union[float, Symbol]:
        """
        Wrapper for LLDB's EvaluateExpression.
        Used for quick code snippets.

        Feel free to use local variables inside the expression using format string.
        For example:
            currentDevice = objc_get_class('UIDevice').currentDevice
            evaluate_expression(f'[[{currentDevice} systemName] hasPrefix:@"2"]')

        :param expression: Expression to evaluate
        :return: Returned value (either float or a Symbol)
        """
        # prepending a prefix so LLDB knows to return an int type
        if isinstance(expression, int):
            formatted_expression = f'(intptr_t)0x{expression:x}'
        else:
            formatted_expression = str(expression)

        options = lldb.SBExpressionOptions()
        options.SetIgnoreBreakpoints(self.configs.evaluation_ignore_breakpoints)
        options.SetTryAllThreads(True)
        options.SetUnwindOnError(self.configs.evaluation_unwind_on_error)

        sbvalue = self.frame.EvaluateExpression(formatted_expression, options)

        if not sbvalue.error.Success():
            raise EvaluatingExpressionError(str(sbvalue.error))

        return self._get_symbol_or_float_from_sbvalue(sbvalue)

    def import_module(self, filename: str, name: Optional[str] = None) -> Any:
        """
        Import & reload given python module (intended mainly for external snippets)
        :param filename: Python filename to import
        :param name: Optional module name, or otherwise use the filename
        :return: Python module
        """
        filename = os.path.expanduser(filename)
        if name is None:
            name = os.path.splitext(os.path.basename(filename))[0]
        spec = importlib.util.spec_from_file_location(name, filename)
        m = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(m)
        return m

    def set_selected_thread(self, idx: Optional[int] = None) -> None:
        if idx is None:
            thread = selection_prompt(self.process.threads)
        else:
            try:
                thread = [t for t in self.process.threads if t.idx == idx][0]
            except IndexError:
                raise InvalidThreadIndexError()
        self.process.SetSelectedThread(thread)

    def unwind(self) -> bool:
        """ Unwind the stack (useful when get_evaluation_unwind() == False) """
        return self.thread.UnwindInnermostExpression().Success()

    @cached_property
    def pid(self) -> int:
        return self.process.GetProcessID()

    @property
    def thread(self):
        """ Current active thread. """
        if self._bp_frame is not None:
            return self._bp_frame.GetThread()
        return self.process.GetSelectedThread()

    @property
    def frame(self):
        """ Current active frame. """
        if self._bp_frame is not None:
            return self._bp_frame
        return self.thread.GetSelectedFrame()

    @contextmanager
    def stopped(self, interval: int = 0):
        """
        Context-Manager for execution while process is stopped.
        If interval is supplied, then if the device is in running state, it will sleep for the interval
        given before and after execution.
        """
        is_running = self.process.GetState() == lldb.eStateRunning

        if is_running:
            self.stop()
            time.sleep(interval)

        try:
            yield
        finally:
            if is_running:
                time.sleep(interval)
                self.cont()

    @contextmanager
    def safe_malloc(self, size: int):
        """
        Context-Manager for allocating a block of memory which is freed afterwards
        :param size:
        :return:
        """
        block = self.symbols.malloc(size)
        if block == 0:
            raise OSError(f'failed to allocate memory of size: {size} bytes')

        try:
            yield block
        finally:
            self.symbols.free(block)

    @contextmanager
    def sync_mode(self):
        """ Context-Manager for execution while LLDB is in sync mode. """
        is_async = self.debugger.GetAsync()
        self.debugger.SetAsync(False)
        try:
            yield
        finally:
            self.debugger.SetAsync(is_async)

    def init_dynamic_environment(self):
        """ Init session-scoped process dynamic dependencies. """
        self.log_debug('init dynamic environment')
        self._dynamic_env_loaded = True

        self.log_debug('disable mach_msg receive errors')
        try:
            CFRunLoopServiceMachPort_hooks.disable_mach_msg_errors()
        except SymbolAbsentError:
            self.log_warning('failed to disable mach_msg errors')

        objc_code = """
        @import ObjectiveC;
        @import Foundation;
        """
        try:
            self.po(objc_code)
        except EvaluatingExpressionError:
            # first time is expected to fail. bug in LLDB?
            pass

    def log_warning(self, message):
        """ Log at warning level """
        self.logger.warning(message)

    def log_debug(self, message):
        """ Log at debug level """
        self.logger.debug(message)

    def log_error(self, message):
        """ Log at error level """
        self.logger.error(message)

    def log_critical(self, message):
        """ Log at critical level """
        self.logger.critical(message)
        raise HildaException(message)

    def log_info(self, message):
        """ Log at info level """
        self.logger.info(message)

    def wait_for_module(self, expression: str) -> None:
        """ Wait for a module to be loaded using `dlopen` by matching given expression """
        self.log_info(f'Waiting for module name containing "{expression}" to be loaded')

        def bp(client: HildaClient, frame, bp_loc, options) -> None:
            loading_module_name = client.evaluate_expression('$arg1').peek_str()
            client.log_info(f'Loading module: {loading_module_name}')
            if expression not in loading_module_name:
                client.cont()
                return
            client.finish()
            client.log_info(f'Desired module has been loaded: {expression}. Process remains stopped')
            bp_id = bp_loc.GetBreakpoint().GetID()
            client.breakpoints.remove(bp_id)

        self.bp('dlopen', bp)
        self.cont()

    def show_help(self, *_) -> None:
        """
        Show banner help message
        """
        help_snippets = [HelpSnippet(key='p', description='Global to access all features')]
        for keybinding in get_keybindings(self):
            help_snippets.append(HelpSnippet(key=keybinding.key.upper(), description=keybinding.description))

        for help_snippet in help_snippets:
            click.echo(help_snippet)

    def interact(self, additional_namespace: Optional[typing.Mapping] = None,
                 startup_files: Optional[list[str]] = None) -> None:
        """ Start an interactive Hilda shell """
        if not self._dynamic_env_loaded:
            self.init_dynamic_environment()

        # Show greeting
        click.secho('Hilda has been successfully loaded! 😎', bold=True)
        click.secho('Usage:', bold=True)
        self.show_help()
        click.echo(click.style('Have a nice flight ✈️! Starting an IPython shell...', bold=True))

        # Configure and start IPython shell
        ipython_config = Config()
        ipython_config.IPCompleter.use_jedi = False
        ipython_config.BaseIPythonApplication.profile = 'hilda'
        ipython_config.InteractiveShellApp.extensions = ['hilda.ipython_extensions.magics',
                                                         'hilda.ipython_extensions.events',
                                                         'hilda.ipython_extensions.keybindings']
        ipython_config.InteractiveShellApp.exec_lines = ['disable_logs()']
        if startup_files is not None:
            ipython_config.InteractiveShellApp.exec_files = startup_files
            self.log_debug(f'Startup files - {startup_files}')

        namespace = self.globals
        namespace['p'] = self
        namespace['ui'] = self.ui_manager
        namespace['cfg'] = self.configs
        if additional_namespace is not None:
            namespace.update(additional_namespace)
        sys.argv = ['a']
        IPython.start_ipython(config=ipython_config, user_ns=namespace)

    def toggle_enable_stdout_stderr(self, *args) -> None:
        self.configs.enable_stdout_stderr = not self.configs.enable_stdout_stderr
        self.logger.info(f'Changed stdout and stderr status to: {self.configs.enable_stdout_stderr}')

    def __enter__(self) -> 'HildaClient':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.detach()

    def _add_global(self, name: str, value: Any, reserved_names=None) -> None:
        if reserved_names is None or name not in reserved_names:
            # don't override existing symbols
            self.globals[name] = value

    @staticmethod
    def _get_saved_state_filename():
        return '/tmp/cache.hilda'

    @staticmethod
    def _to_ns_json_default(obj):
        if isinstance(obj, bytes):
            return f'__hilda_magic_key__|NSData|{base64.b64encode(obj).decode()}'
        elif isinstance(obj, datetime):
            return f'__hilda_magic_key__|NSDate|{obj.timestamp()}'
        raise TypeError

    @staticmethod
    def _from_ns_json_object_hook(obj: dict):
        parsed_object = {}
        for key, value in obj.items():
            parsed_object[HildaClient._from_ns_parse_function(key)] = HildaClient._from_ns_parse_function(value)
        return parsed_object

    @staticmethod
    def _from_ns_parse_function(obj):
        if isinstance(obj, list):
            return list(map(HildaClient._from_ns_parse_function, obj))
        if not isinstance(obj, str) or not obj.startswith('__hilda_magic_key__'):
            return obj
        _, type_, data = obj.split('|')
        if type_ == 'NSData':
            return base64.b64decode(data)
        if type_ == 'NSDictionary':
            return tuple(json.loads(data, object_hook=HildaClient._from_ns_json_object_hook).items())
        if type_ == 'NSArray':
            return tuple(json.loads(data, object_hook=HildaClient._from_ns_json_object_hook))
        if type_ == 'NSNumber':
            return eval(data)
        if type_ == 'NSNull':
            return None
        if type_ == 'NSDate':
            return datetime.fromtimestamp(eval(data), timezone.utc)

    def _serialize_call_params(self, argv):
        args_conv = []
        for arg in argv:
            if isinstance(arg, str) or isinstance(arg, bytes):
                if isinstance(arg, str):
                    arg = arg.encode()
                arg = ''.join([f'\\x{b:02x}' for b in arg])
                args_conv.append(f'(intptr_t)"{arg}"')
            elif isinstance(arg, int) or isinstance(arg, Symbol):
                arg = int(arg) & 0xffffffffffffffff
                args_conv.append(f'0x{arg:x}')
            else:
                raise NotImplementedError('cannot serialize argument')
        return args_conv

    def _generate_call_expression(self, address, params):
        args_type = ','.join(['intptr_t'] * len(params))
        args_conv = ','.join(params)

        if self.arch == 'arm64e':
            address = f'ptrauth_sign_unauthenticated((void *){address}, ptrauth_key_asia, 0)'

        return f'((intptr_t(*)({args_type}))({address}))({args_conv})'

    @staticmethod
    def _std_string(value: Symbol) -> str:
        if struct.unpack("b", (value + 23).peek(1))[0] >= 0:
            return value.peek_str()
        else:
            return value[0].peek_str()

    @cached_property
    def _object_identifier(self) -> Symbol:
        return self.symbols.objc_getClass('VMUObjectIdentifier').objc_call('alloc').objc_call(
            'initWithTask:', self.symbols.mach_task_self())

    @cached_property
    def _ks(self) -> Optional['Ks']:
        if not lldb.KEYSTONE_SUPPORT:
            return False
        platforms = {'arm64': Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN),
                     'arm64e': Ks(KS_ARCH_ARM64, KS_MODE_LITTLE_ENDIAN),
                     'x86_64h': Ks(KS_ARCH_X86, KS_MODE_64)}
        return platforms.get(self.arch)

    def _get_module_class_list(self, module_name: str):
        for m in self.target.module_iter():
            if module_name != m.file.basename:
                continue
            objc_classlist = m.FindSection('__DATA').FindSubSection('__objc_classlist')
            objc_classlist_addr = self.symbol(objc_classlist.GetLoadAddress(self.target))
            obj_c_code = (self._hilda_root / 'objective_c' / 'get_objectivec_class_by_module.m').read_text()
            obj_c_code = obj_c_code.replace('__count_objc_class', f'{objc_classlist.size // 8}').replace(
                '__objc_class_list',
                f'{objc_classlist_addr}')
            return json.loads(self.po(obj_c_code))

    def _get_symbol_or_float_from_sbvalue(self, value: lldb.SBValue) -> Union[float, Symbol]:
        # The `value` attribute of an SBValue stores a string representation of the actual value
        # in a python-compatible format, so we can eval it to get the native python value
        value = eval(value.value)
        if isinstance(value, float):
            return value
        return self.symbol(value)



from hilda.breakpoints import WhereType
from hilda.lldb_importer import lldb

p = lldb.hilda_client


def breakpoint(where: WhereType, enable: bool = False):
    """
    A decorator to define a breakpoint, e.g.,

    ```
    @breakpoint('symbol_name')
    def my_breakpoint(hilda, *args):
        print('Hit!')
        hilda.cont()

    my_breakpoint.enabled = True
    ```
    """
    def decorator(callback):
        bp = p.breakpoints.add(where, callback=callback)
        bp.enabled = enable
        return bp

    return decorator


def watchpoint(where: int, enable: bool = False):
    """
    A decorator to define a watchpoint, e.g.,

    ```
    @watchpoint(0x11223344)
    def my_watchpoint(hilda, *args):
        print('Hit!')
        hilda.cont()

    my_watchpoint.enabled = True
    ```
    """
    def decorator(callback):
        wp = p.watchpoints.add(where, callback=callback)
        wp.enabled = enable
        return wp

    return decorator



from hilda.exceptions import SymbolAbsentError
from hilda.lldb_importer import lldb


def _CFRunLoopServiceMachPort_hook(hilda, *args):
    """
    :param hilda.hilda_client.HildaClient hilda:
     """
    hilda.jump(hilda.CFRunLoopServiceMachPort_while_ea)
    hilda.cont()


def _disable_internal_error_handling() -> None:
    hilda = lldb.hilda_client
    with hilda.stopped():
        instructions = hilda.symbols.__CFRunLoopServiceMachPort.disass(2000, should_print=False)
        while_ea = None
        for instruction in instructions:
            if (while_ea is None) and instruction.DoesBranch():
                # Beginning of the `while(true) { ... }`
                while_ea = instruction.GetOperands(hilda.target)
                hilda.CFRunLoopServiceMachPort_while_ea = int(hilda.file_symbol(eval(while_ea)))
            elif instruction.GetMnemonic(hilda.target) in ('brk', 'ud2'):
                symbol = hilda.symbol(instruction.addr.GetLoadAddress(hilda.target))
                description = f'__CFRunLoopServiceMachPort-brk-{int(symbol - hilda.symbols.__CFRunLoopServiceMachPort)}'
                symbol.bp(_CFRunLoopServiceMachPort_hook, guarded=True, description=description)

    if hilda.arch == 'x86_64h':
        return

    # on iOS 16.x, will need to also patch this one
    try:
        handle_error = hilda.symbols['__CFRunLoopServiceMachPort.cold.1']
    except SymbolAbsentError:
        return

    for instruction in handle_error.disass(2000, should_print=False):
        if instruction.GetMnemonic(hilda.target) in ('brk', 'ud2'):
            # mov x0, x0
            hilda.symbol(instruction.addr.GetLoadAddress(hilda.target)).poke(b'\xe0\x03\x00\xaa')


def _disable_mach_msg_timeout() -> None:
    """
    Remove the timeout validation from __CFRunLoopServiceMachPort. This is done by patching the mach_msg_timeout_t
    parameter (4rd one) to MACH_MSG_TIMEOUT_NONE. On arm the parameter passes on `x3` register.

        __int64 __fastcall __CFRunLoopServiceMachPort(
            mach_port_name_t a1,
            mach_msg_header_t **a2,
            mach_port_t *a3,
            mach_msg_timeout_t a4,
            voucher_mach_msg_state_t *a5,
            id *a6)

        CoreFoundation:__text:0000000186E5012C ___CFRunLoopServiceMachPort
        CoreFoundation:__text:0000000186E5012C                 SUB             SP, SP, #0x70
        CoreFoundation:__text:0000000186E50130                 STP             X28, X27, [SP,#0x60+var_50]
        CoreFoundation:__text:0000000186E50134                 STP             X26, X25, [SP,#0x60+var_40]
        CoreFoundation:__text:0000000186E50138                 STP             X24, X23, [SP,#0x60+var_30]
        CoreFoundation:__text:0000000186E5013C                 STP             X22, X21, [SP,#0x60+var_20]
        CoreFoundation:__text:0000000186E50140                 STP             X20, X19, [SP,#0x60+var_10]
        CoreFoundation:__text:0000000186E50144                 STP             X29, X30, [SP,#0x60+var_s0]
        CoreFoundation:__text:0000000186E50148                 ADD             X29, SP, #0x60
        CoreFoundation:__text:0000000186E5014C                 MOV             X21, X5
        CoreFoundation:__text:0000000186E50150                 MOV             X22, X4
        CoreFoundation:__text:0000000186E50154                 MOV             X23, X3       <-------- Timeout parameter
    """
    if not lldb.KEYSTONE_SUPPORT:
        return

    hilda = lldb.hilda_client
    if hilda.arch == 'x86_64h':
        return

    with hilda.stopped():
        for inst in hilda.symbols.__CFRunLoopServiceMachPort.disass(200, should_print=False):
            mnemonic = inst.GetMnemonic(hilda.target)
            operands = inst.GetOperands(hilda.target)
            if mnemonic != 'mov' or not operands.endswith('x3'):
                continue
            addr = inst.GetAddress()
            file_addr = addr.GetFileAddress()
            new_inst = f'{mnemonic} {operands.replace("x3", "0")}'
            hilda.file_symbol(file_addr).poke_text(new_inst)
            break


def disable_mach_msg_errors() -> None:
    _disable_mach_msg_timeout()
    _disable_internal_error_handling()



from hilda.objective_c_symbol import ObjectiveCSymbol


def iter_values(symbol: ObjectiveCSymbol):
    enum = symbol.objectEnumerator()
    item = enum.nextObject()
    while item != 0:
        yield item
        item = enum.nextObject()


def iter_keys(symbol: ObjectiveCSymbol):
    enum = symbol.keyEnumerator()
    item = enum.nextObject()
    while item != 0:
        yield item
        item = enum.nextObject()



import json
import time
from collections import UserList, namedtuple
from dataclasses import dataclass, field
from functools import partial
from pathlib import Path
from typing import Any, Optional

from objc_types_decoder.decode import decode as decode_type
from objc_types_decoder.decode import decode_with_tail
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from hilda.exceptions import GettingObjectiveCClassError

Ivar = namedtuple('Ivar', 'name type_ offset')
Property = namedtuple('Property', 'name attributes')
PropertyAttributes = namedtuple('PropertyAttributes', 'synthesize type_ list')


def convert_encoded_property_attributes(encoded):
    conversions = {
        'R': lambda x: 'readonly',
        'C': lambda x: 'copy',
        '&': lambda x: 'strong',
        'N': lambda x: 'nonatomic',
        'G': lambda x: 'getter=' + x[1:],
        'S': lambda x: 'setter=' + x[1:],
        'd': lambda x: 'dynamic',
        'W': lambda x: 'weak',
        'P': lambda x: '<garbage-collected>',
        't': lambda x: 'encoding=' + x[1:],
    }

    type_, tail = decode_with_tail(encoded[1:])
    attributes = []
    synthesize = None
    for attr in filter(None, tail.lstrip(',').split(',')):
        if attr[0] in conversions:
            attributes.append(conversions[attr[0]](attr))
        elif attr[0] == 'V':
            synthesize = attr[1:]

    return PropertyAttributes(type_=type_, synthesize=synthesize, list=attributes)


@dataclass
class Method:
    name: str
    client: Any = field(compare=False)
    address: int = field(compare=False)
    imp: int = field(compare=False)
    type_: str = field(compare=False)
    return_type: str = field(compare=False)
    is_class: bool = field(compare=False)
    args_types: list = field(compare=False)

    @staticmethod
    def from_data(data: dict, client):
        """
        Create Method object from raw data.
        :param data: Data as loaded from get_objectivec_symbol_data.m.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        """
        return Method(
            name=data['name'],
            client=client,
            address=client.symbol(data['address']),
            imp=client.symbol(data['imp']),
            type_=data['type'],
            return_type=decode_type(data['return_type']),
            is_class=data['is_class'],
            args_types=list(map(decode_type, data['args_types']))
        )

    def set_implementation(self, new_imp: int):
        self.client.symbols.method_setImplementation(self.address, new_imp)
        self.imp = self.client.symbol(new_imp)

    def __str__(self):
        if ':' in self.name:
            args_names = self.name.split(':')
            name = ' '.join(['{}:({})'.format(*arg) for arg in zip(args_names, self.args_types[2:])])
        else:
            name = self.name
        prefix = '+' if self.is_class else '-'
        return f'{prefix} {name}; // 0x{self.address:x} (returns: {self.return_type})\n'


class MethodList(UserList):
    def __init__(self, class_name: str, methods: list[Method]) -> None:
        super().__init__()
        self._class_name = class_name
        self.data = methods

    def get(self, name: str) -> Optional[Method]:
        """
        Get a specific method implementation.
        :param name: Method name.
        :return: Method.
        """
        for method in self.data:
            if method.name == name:
                return method
        return None

    # Actions

    def bp(self, callback=None, **kwargs):
        """
        Place a breakpoint on all symbols in the method list.
        Look for the bp command for more details.
        :param callback:  callback function to be executed upon an hit
        :param kwargs: optional kwargs for the bp command
        """
        for method in self.data:
            kwargs['name'] = f'[{self._class_name} {method.name}]'
            method.imp.bp(callback, **kwargs)

    def monitor(self, **kwargs):
        """
        Perform monitor for all symbols in the method list.
        See monitor command for more details.
        :param args: given arguments for monitor command
        """
        for method in self.data:
            method.imp.monitor(**kwargs)

    # Filters

    def filter_startswith(self, exp, case_sensitive=True):
        """
        Filter only methods with given prefix
        :param exp: prefix
        :param case_sensitive: is case sensitive
        :return: reduced method list
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = []
        for v in self.data:
            name = v.name
            if not case_sensitive:
                name = name.lower()
            if name.startswith(exp):
                retval.append(v)
        return MethodList(self._class_name, retval)

    def filter_name_contains(self, exp, case_sensitive=True):
        """
        Filter methods containing a given expression
        :param exp: given expression
        :param case_sensitive: is case sensitive
        :return: reduced method list
        """
        if not case_sensitive:
            exp = exp.lower()

        retval = []
        for v in self.data:
            name = v.name
            if not case_sensitive:
                name = name.lower()
            if exp in name:
                retval.append(v)
        return MethodList(self._class_name, retval)


class Class:
    """
    Wrapper for ObjectiveC Class object.
    """

    def __init__(self, client, class_object=0, class_data: dict = None):
        """
        :param hilda.hilda_client.HildaClient client:
        :param hilda.symbol.Symbol class_object:
        """
        self._client = client
        self._class_object = class_object
        self.protocols = []
        self.ivars = []
        self.properties = []
        self.name = ''
        self.methods = MethodList(self.name, [])
        self.super = None
        if class_data is None:
            self.reload()
        else:
            self._load_class_data(class_data)

    @staticmethod
    def from_class_name(client, class_name: str):
        """
        Create ObjectiveC Class from given class name.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        :param class_name: Class name.
        """
        obj_c_code = (client._hilda_root / 'objective_c' / 'get_objectivec_class_description.m').read_text()
        obj_c_code = obj_c_code.replace('__class_address__', '0').replace('__class_name__', class_name)
        class_symbol = Class(client, class_data=json.loads(client.po(obj_c_code)))
        if class_symbol.name != class_name:
            raise GettingObjectiveCClassError()
        return class_symbol

    @staticmethod
    def sanitize_name(name: str):
        """
        Sanitize python name to ObjectiveC name.
        """
        if name.startswith('_'):
            name = '_' + name[1:].replace('_', ':')
        else:
            name = name.replace('_', ':')
        return name

    def reload(self):
        """
        Reload class object data.
        Should be used whenever the class layout changes (for example, during method swizzling)
        """
        obj_c_code = (self._client._hilda_root / 'objective_c' / 'get_objectivec_class_description.m').read_text()
        obj_c_code = obj_c_code.replace('__class_address__', f'{self._class_object:d}')
        obj_c_code = obj_c_code.replace('__class_name__', self.name)
        self._load_class_data(json.loads(self._client.po(obj_c_code)))

    def show(self):
        """
        Print to terminal the highlighted class description.
        """
        print(highlight(str(self), ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

    def objc_call(self, sel: str, *args):
        """
        Invoke a selector on the given class object.
        :param sel: Selector name.
        :return: whatever the selector returned as a symbol.
        """
        return self._class_object.objc_call(sel, *args)

    def capture_self(self, sync: bool = False):
        """
        Capture the first called `self` from this class.
        Access using `self.captured_objects`
        :param sync: Should wait until captured object is returned?
        :return: Captured object if sync is True, None otherwise
        """
        class_name = self.name

        if class_name in self._client.captured_objects:
            del self._client.captured_objects[class_name]

        def hook(hilda, frame, bp_loc, options):
            hilda.log_info(f'self object has been captured from {options["name"]}')
            hilda.log_info('removing breakpoints')
            for bp_id, bp in list(hilda.breakpoints.items()):
                if 'group_uuid' in bp.options and bp.options.get('group_uuid', '') == options['group_uuid']:
                    hilda.breakpoints.remove(bp_id)
            captured = hilda.evaluate_expression('$arg1')
            captured = captured.objc_symbol
            hilda.captured_objects[options['name'].split(' ')[0].split('[')[1]] = captured
            hilda.cont()

        for method in self.methods:
            if not method.is_class:
                # only instance methods are relevant for capturing self
                method.imp.bp(hook)

        if sync:
            self._client.cont()
            self._client.log_debug('Waiting for desired object to be captured...')
            while class_name not in self._client.captured_objects:
                time.sleep(1)

            return self._client.captured_objects[class_name]

    def monitor(self, **kwargs):
        """
        Proxy for monitor command.
        """
        self.methods.monitor(**kwargs)

    def bp(self, callback=None, **kwargs):
        """
        Proxy for bp command.
        """
        self.methods.bp(callback, **kwargs)

    def iter_supers(self):
        """
        Iterate over the super classes of the class.
        """
        sup = self.super
        while sup is not None:
            yield sup
            sup = sup.super

    @property
    def bundle_path(self) -> Path:
        return Path(self._client.symbols.objc_getClass('NSBundle')
                    .objc_call('bundleForClass:', self._class_object).objc_call('bundlePath').py())

    def _load_class_data(self, data: dict):
        self._class_object = self._client.symbol(data['address'])
        self.super = Class(self._client, data['super']) if data['super'] else None
        self.name = data['name']
        self.protocols = data['protocols']
        self.ivars = [
            Ivar(name=ivar['name'], type_=decode_type(ivar['type']) if ivar['type'] else 'unknown_type_t',
                 offset=ivar['offset'])
            for ivar in data['ivars']
        ]
        self.properties = [
            Property(name=prop['name'], attributes=convert_encoded_property_attributes(prop['attributes']))
            for prop in data['properties']
        ]
        self.methods = MethodList(self.name, [Method.from_data(method, self._client) for method in data['methods']])

    def __dir__(self):
        result = set()

        for method in self.methods:
            if method.is_class:
                result.add(method.name.replace(':', '_'))

        for sup in self.iter_supers():
            if self._client.configs.nsobject_exclusion and sup.name == 'NSObject':
                continue
            for method in sup.methods:
                if method.is_class:
                    result.add(method.name.replace(':', '_'))

        result.update(list(super().__dir__()))
        return list(result)

    def __str__(self):
        protocol_buf = f'<{",".join(self.protocols)}>' if self.protocols else ''

        if self.super is not None:
            buf = f'@interface {self.name}: {self.super.name} {protocol_buf}\n'
        else:
            buf = f'@interface {self.name} {protocol_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            buf += f'@property ({",".join(prop.attributes.list)}) {prop.attributes.type_} {prop.name};\n'

            if prop.attributes.synthesize is not None:
                buf += f'@synthesize {prop.name} = {prop.attributes.synthesize};\n'

        # Add methods
        for method in self.methods:
            buf += str(method)

        buf += '@end'
        return buf

    def __repr__(self):
        return f'<objC Class "{self.name}">'

    def __getitem__(self, item):
        for method in self.methods:
            if method.name == item:
                if method.is_class:
                    return partial(self.objc_call, item)
                else:
                    raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                         f'not a class method')

        for sup in self.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    if method.is_class:
                        return partial(self.objc_call, item)
                    else:
                        raise AttributeError(f'{self.name} class has an instance method named {item}, '
                                             f'not a class method')

        raise AttributeError(f''''{self.name}' class has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.sanitize_name(item)]



class Registers:
    """
    Wrapper for more convenient access to modify current frame's registers
    """

    def __init__(self, client):
        self.__dict__['_client'] = client

    def __getattr__(self, item):
        return self._client.get_register(item)

    def __getitem__(self, item):
        return self._client.get_register(item)

    def __setattr__(self, key, value):
        return self._client.set_register(key, value)

    def __setitem__(self, key, value):
        return self._client.set_register(key, value)

    def __dir__(self):
        result = []
        for group in self._client.frame.register.regs:
            for register in group:
                result.append(register.name)
        return result

    def show(self):
        """ Show current frame's registers """
        print(self._client.frame.register.regs)



import pytest

from hilda.exceptions import AccessingRegisterError


@pytest.mark.parametrize('register_name', ['x0', 'x28', 'w0', 'w28', 'sp', 'lr', 'pc'])
def test_register_dir(hilda_client, register_name: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param register_name: Register that suppose to appear in dir.
    """
    assert register_name in dir(hilda_client.registers)


def test_register_get(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    for i in range(29):
        register_name = f'x{i}'
        assert hilda_client.registers[register_name] == hilda_client.evaluate_expression(f'${register_name}')
        assert getattr(hilda_client.registers, register_name) == hilda_client.registers[register_name]

    for i in range(32):
        register_name = f'd{i}'
        assert hilda_client.registers[register_name] == hilda_client.evaluate_expression(f'${register_name}')
        assert getattr(hilda_client.registers, register_name) == hilda_client.registers[register_name]


def test_register_get_uppercase(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert hilda_client.registers.x0 == hilda_client.registers.X0
    assert hilda_client.registers.pc == hilda_client.registers.PC


def test_register_set(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    original_x0 = hilda_client.registers.x0
    hilda_client.registers.x0 = 1337
    assert hilda_client.registers.x0 == 1337
    hilda_client.registers['x0'] = 1338
    assert hilda_client.registers.x0 == 1338
    hilda_client.registers.x0 = original_x0
    hilda_client.registers.d0 = 133.7
    assert hilda_client.registers.d0 == 133.7


def test_register_set_uppercase(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    original_x0 = hilda_client.registers.x0
    hilda_client.registers['X0'] = 1339
    assert hilda_client.registers.x0 == 1339
    hilda_client.registers.x0 = original_x0


def test_get_register_error(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(AccessingRegisterError):
        hilda_client.registers.registerthatdoesntexist


def test_set_register_error(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(AccessingRegisterError):
        hilda_client.registers.registerthatdoesntexist = 2



import json
from contextlib import suppress
from dataclasses import dataclass
from functools import partial

from objc_types_decoder.decode import decode as decode_type
from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from hilda.exceptions import HildaException
from hilda.objective_c_class import Class, Method, MethodList, Property, convert_encoded_property_attributes
from hilda.symbol import Symbol


class SettingIvarError(HildaException):
    """ Raise when trying to set an Ivar too early or when the Ivar doesn't exist. """
    pass


@dataclass
class Ivar:
    name: str
    value: Symbol
    type_: str
    offset: int


class ObjectiveCSymbol(Symbol):
    """
    Wrapper object for an objective-c symbol.
    Allowing easier access to its properties, methods and ivars.
    """

    @classmethod
    def create(cls, value: int, client):
        """
        Create an ObjectiveCSymbol object.
        :param value: Symbol address.
        :param hilda.hilda_client.HildaClient client: hilda client.
        :return: ObjectiveCSymbol object.
        :rtype: ObjectiveCSymbol
        """
        symbol = super().create(value, client)
        symbol.ivars = []  # type: List[Ivar]
        symbol.properties = []  # type: List[Property]
        symbol.methods = MethodList('', [])  # type: MethodList
        symbol.class_ = None  # type: Optional[Class]
        symbol.reload()
        return symbol

    def reload(self):
        """
        Reload object's in-memory layout.
        """
        self.ivars.clear()
        self.properties.clear()
        self.methods = MethodList('', [])
        self.class_ = None

        obj_c_code = (self._client._hilda_root / 'objective_c' / 'get_objectivec_symbol_data.m').read_text()
        obj_c_code = obj_c_code.replace('__symbol_address__', f'{self:d}')
        data = json.loads(self._client.po(obj_c_code))

        self._reload_ivars(data['ivars'])
        self._reload_properties(data['properties'])

        data['name'] = data['class_name']
        data['address'] = data['class_address']
        data['super'] = data['class_super']
        self.class_ = Class(self._client, self._client.symbol(data['class_address']), data)
        self.methods = MethodList(self.class_.name, [Method.from_data(method, self._client) for method in data['methods']])

    def show(self, recursive: bool = False):
        """
        Print to terminal the highlighted class description.
        :param recursive: Show methods of super classes.
        """
        print(highlight(self._to_str(recursive), ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

    def _reload_ivars(self, ivars_data):
        raw_ivars = sorted(ivars_data, key=lambda ivar: ivar['offset'])
        for i, ivar in enumerate(raw_ivars):
            ivar_type = ivar['type']
            if ivar_type:
                ivar_type = decode_type(ivar_type)
            else:
                ivar_type = 'unknown_type_t'
            value = ivar['value']
            if i < len(raw_ivars) - 1:
                # The .fm file returns a 64bit value, regardless of the real size.
                size = raw_ivars[i + 1]['offset'] - ivar['offset']
                value = value & ((2 ** (size * 8)) - 1)
            ivar_value = self._client.symbol(value)
            self.ivars.append(Ivar(name=ivar['name'], type_=ivar_type, offset=ivar['offset'], value=ivar_value))

    def _reload_properties(self, properties_data):
        for prop in properties_data:
            prop_attributes = convert_encoded_property_attributes(prop['attributes'])
            self.properties.append(Property(name=prop['name'], attributes=prop_attributes))

    def _set_ivar(self, name, value):
        try:
            ivars = self.__getattribute__('ivars')
            class_name = self.__getattribute__('class_').name
        except AttributeError as e:
            raise SettingIvarError from e

        for i, ivar in enumerate(ivars):
            if ivar.name == name:
                size = self.item_size
                if i < len(self.ivars) - 1:
                    size = ivars[i + 1].offset - ivar.offset
                with self.change_item_size(size):
                    self[ivar.offset // size] = value
                    ivar.value = value
                return
        raise SettingIvarError(f'Ivar "{name}" does not exist in "{class_name}"')

    def _to_str(self, recursive=False):
        protocols_buf = f'<{",".join(self.class_.protocols)}>' if self.class_.protocols else ''

        if self.class_.super is not None:
            buf = f'@interface {self.class_.name}: {self.class_.super.name} {protocols_buf}\n'
        else:
            buf = f'@interface {self.class_.name} {protocols_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name} = 0x{int(ivar.value):x}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            attrs = prop.attributes
            buf += f'@property ({",".join(attrs.list)}) {prop.attributes.type_} {prop.name};\n'

            if attrs.synthesize is not None:
                buf += f'@synthesize {prop.name} = {attrs.synthesize};\n'

        # Add methods
        methods = list(self.methods)

        # Add super methods.
        if recursive:
            for sup in self.class_.iter_supers():
                for method in filter(lambda m: m not in methods, sup.methods):
                    methods.append(method)

        # Print class methods first.
        methods.sort(key=lambda m: not m.is_class)

        for method in methods:
            buf += str(method)

        buf += '@end'
        return buf

    def __dir__(self):
        result = set()

        for ivar in self.ivars:
            result.add(ivar.name)

        for method in self.methods:
            result.add(method.name.replace(':', '_'))

        for sup in self.class_.iter_supers():
            if self._client.configs.nsobject_exclusion and sup.name == 'NSObject':
                continue
            for method in sup.methods:
                result.add(method.name.replace(':', '_'))

        result.update(list(super().__dir__()))
        return list(result)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super().__getitem__(item)

        # Ivars
        for ivar in self.ivars:
            if ivar.name == item:
                return ivar.value

        # Properties
        for prop in self.properties:
            if prop.name == item:
                return self.objc_call(item)

        # Methods
        for method in self.methods:
            if method.name == item:
                return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        for sup in self.class_.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        raise AttributeError(f''''{self.class_.name}' has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.class_.sanitize_name(item)]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            super().__setitem__(key, value)
            return

        with suppress(SettingIvarError):
            self._set_ivar(key, value)
            return

    def __setattr__(self, key, value):
        try:
            key = self.__getattribute__('class_').sanitize_name(key)
        except AttributeError:
            pass
        try:
            self._set_ivar(key, value)
        except SettingIvarError:
            super().__setattr__(key, value)

    def __str__(self):
        return self._to_str(False)

    def __repr__(self):
        return f'<{self.__class__.__name__} 0x{int(self):x} Class: {self.class_.name}>'



import logging
from pathlib import Path
from typing import Optional

import click
import coloredlogs

from hilda import launch_lldb
from hilda._version import version
from hilda.launch_lldb import create_hilda_client_using_attach_by_name, create_hilda_client_using_attach_by_pid, \
    create_hilda_client_using_launch, create_hilda_client_using_remote_attach

DEFAULT_HILDA_PORT = 1234

coloredlogs.install(level=logging.DEBUG)


@click.group()
def cli():
    pass


startup_files_option = click.option('-f', '--startup_files', multiple=True, help='Files to run on start')


def parse_envp(ctx: click.Context, param: click.Parameter, value: list[str]) -> list[str]:
    env_list = []
    for item in value:
        try:
            key, val = item.split('=', 1)
            env_list.append(f'{key}={val}')
        except ValueError:
            raise click.BadParameter(f'Invalid format for --envp: {item}. Expected KEY=VALUE.')
    return env_list


@cli.command('remote')
@click.argument('hostname', default='localhost')
@click.argument('port', type=click.INT, default=DEFAULT_HILDA_PORT)
@startup_files_option
def remote(hostname: str, port: int, startup_files: list[str]) -> None:
    """ Connect to remote debugserver at given address """
    with create_hilda_client_using_remote_attach(hostname, port) as hilda_client:
        hilda_client.interact(startup_files=startup_files)


@cli.command('attach')
@click.option('-n', '--name', help='process name to attach')
@click.option('-p', '--pid', type=click.INT, help='pid to attach')
@startup_files_option
def attach(name: Optional[str], pid: Optional[int], startup_files: list[str]) -> None:
    """ Attach to given process and start a lldb shell """
    if name is not None:
        hilda_client = create_hilda_client_using_attach_by_name(name)
    elif pid is not None:
        hilda_client = create_hilda_client_using_attach_by_pid(pid)
    else:
        raise click.UsageError('You must specify a process name or pid')
    hilda_client.interact(startup_files=startup_files)
    hilda_client.detach()


@cli.command('bare')
def cli_bare() -> None:
    """ Just start a lldb shell """
    commands = [f'command script import {Path(__file__).resolve().parent / "lldb_entrypoint.py"}']
    commands = '\n'.join(commands)
    launch_lldb.execute(f'lldb --one-line "{commands}"')


@cli.command('launch')
@click.argument('exec_path')
@click.option('--argv', multiple=True, help='Command line arguments to pass to the process')
@click.option('--envp', multiple=True, callback=parse_envp, help='Environment variables in the form KEY=VALUE')
@click.option('--stdin', help='Redirect stdin from this file path')
@click.option('--stdout', help='Redirect stdout to this file path')
@click.option('--stderr', help='Redirect stderr to this file path')
@click.option('--cwd', help='Set the working directory for the process')
@click.option('--flags', type=click.INT, default=0, help='Launch flags (bitmask)')
@startup_files_option
def launch(exec_path: str, argv: list[str], envp: list[str], stdin: Optional[Path],
           stdout: Optional[Path], stderr: Optional[Path], cwd: Optional[Path], flags: Optional[int],
           startup_files: list[str]) -> None:
    """ Attach to a given process and start a lldb shell """
    argv = list(argv)
    envp = list(envp)
    with create_hilda_client_using_launch(
            exec_path, argv, envp, stdin, stdout, stderr, cwd, flags) as hilda_client:
        hilda_client.interact(startup_files=startup_files)


@cli.command('version')
def cli_version() -> None:
    """Show the version information."""
    click.echo(version)



import pytest

from hilda.exceptions import CreatingObjectiveCSymbolError
from hilda.objective_c_symbol import ObjectiveCSymbol


@pytest.mark.parametrize('literal, result', [
    ('@{@"one": @1}', '{\n    one = 1;\n}'),
    ('@{}', '{\n}'),
])
def test_create_objective_c_symbol(hilda_client, literal: str, result: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param literal: ObjectiveC object literal.
    :param result: Object po.
    """
    assert hilda_client.evaluate_expression(literal).objc_symbol.po() == result


def test_fail_create_objective_c_symbol(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(CreatingObjectiveCSymbolError):
        hilda_client.symbol(1).objc_symbol


def test_access_missing_attribute(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    with pytest.raises(AttributeError):
        assert single_key_dictionary.attribute_that_doesnt_exist


def test_get_objective_c_ivar(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    assert single_key_dictionary._key.po() == 'one'


def test_change_objective_c_ivar(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    single_key_dictionary._key = hilda_client.ns('two')
    assert single_key_dictionary.description.po() == '{\n    two = 1;\n}'
    assert single_key_dictionary._key.po() == 'two'


def test_call_objective_c_property(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    single_key_dictionary = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    assert single_key_dictionary.description.po() == '{\n    one = 1;\n}'


def test_call_objective_c_method_by_method_name(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    dictionary1 = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    dictionary2 = hilda_client.evaluate_expression('@{@"one": @2}').objc_symbol
    assert dictionary1.isEqualToDictionary_(dictionary1)
    assert not dictionary1.isEqualToDictionary_(dictionary2)


def test_call_objective_c_method_by_objc_call(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    dictionary1 = hilda_client.evaluate_expression('@{@"one": @1}').objc_symbol
    dictionary2 = hilda_client.evaluate_expression('@{@"one": @2}').objc_symbol
    assert dictionary1.objc_call('isEqualToDictionary:', dictionary1)
    assert not dictionary1.objc_call('isEqualToDictionary:', dictionary2)


def test_call_objective_c_method_returns_native_symbol(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    string1 = hilda_client.ns('asdasd').objc_symbol
    assert not isinstance(string1.UTF8String(), ObjectiveCSymbol)


@pytest.mark.parametrize('selector', [
    'alloc',  # Test super class methods.
    'init',  # Test super instance methods.
    'dictionaryWithObjects_forKeys_',  # Test class methods.
    'objectForKey_',  # Test instance methods.
    'allKeys',  # Test instance properties.
])
def test_symbol_dir(hilda_client, selector: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param selector: Superclass selector to test.
    """
    dict_dir = dir(hilda_client.ns({1: 2}).objc_symbol)
    assert dict_dir.count(selector) == 1


@pytest.mark.parametrize('sub_str', [
    '@interface NSObject',
    '+ alloc;',
    'Class isa = 0x',
    '@property (readonly,copy) NSString * description;',
])
def test_symbol_without_super_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert str(hilda_client.objc_get_class('NSObject').new().objc_symbol).count(sub_str) == 1


@pytest.mark.parametrize('sub_str', [
    '@interface',  # Class declaration.
    'NSMutableDictionary',  # Class type.
    'Class isa = 0x',  # Ivar.
    '@property (readonly,copy) NSString * description;',  # Property.
    '+ automaticallyNotifiesObserversForKey:(id);',  # Class method.
    '- removeAllObjects;',  # Instance method.
])
def test_symbol_str(hilda_client, sub_str: str):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    :param sub_str: Substring that should appear in the str.
    """
    assert hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str().count(sub_str) == 1


def test_symbol_str_not_recursive(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert '+ alloc;' not in hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str(False)


def test_symbol_str_recursive(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    assert '+ alloc;' in hilda_client.ns({1: 2, 3: 4}).objc_symbol._to_str(True)


@pytest.mark.skip('Not fully supported')
def test_set_implementation(hilda_client):
    pid = hilda_client.symbols.getpid()

    hilda_client.objc_get_class('NSJSONSerialization').methods.get('isValidJSONObject:').set_implementation(
        hilda_client.symbols.getpid)
    assert hilda_client.objc_get_class('NSJSONSerialization').isValidJSONObject_() == pid



import os
import struct
from contextlib import contextmanager
from functools import cached_property
from typing import Any, Optional, Tuple

from construct import FormatField

from hilda.common import CfSerializable
from hilda.lldb_importer import lldb
from hilda.objective_c_class import Class

ADDRESS_SIZE_TO_STRUCT_FORMAT = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}


class SymbolFormatField(FormatField):
    """
    A Symbol wrapper for construct
    """

    def __init__(self, client):
        super().__init__('<', 'Q')
        self._client = client

    def _parse(self, stream, context, path):
        return self._client.symbol(FormatField._parse(self, stream, context, path))


"""
A value identifying a `HildaSymbol`.

`HildaSymbol`s are either regular (i.e., named) symbols or anonymous symbols.

Regular symbols are uniquely identified by a `HildaSymbolId` (i.e., no two
instances of `HildaSymbol` have the same ID).
Note that several regular symbols may have the same address (with different names).

Anonymous symbols are not uniquely identified by a `HildaSymbolId`.
"""
HildaSymbolId = Tuple[Optional[str], int]


class Symbol(int):
    """
    Hilda's class representing a symbol (not necessarily an LLDB symbol).
    """

    PROXY_METHODS = ['peek', 'poke', 'peek_str', 'monitor', 'bp',
                     'disass', 'po', 'objc_call']

    @classmethod
    def create(cls, value: int, client, lldb_symbol: Optional[lldb.SBSymbol] = None,
               lldb_address: Optional[lldb.SBAddress] = None, lldb_type: Optional[int] = None) -> None:
        """
        Create a Symbol object.
        :param value: Symbol address.
        :param hilda.hilda_client.HildaClient client: Hilda client.
        :param lldb.SBSymbol lldb_symbol: LLDB symbol.
        :return: Symbol object.
        :rtype: Symbol
        """
        if not isinstance(value, int):
            raise TypeError()

        value &= 0xFFFFFFFFFFFFFFFF

        symbol = cls(value)

        # public properties
        symbol.retval_bit_count = client.RETVAL_BIT_COUNT
        symbol.is_retval_signed = True
        symbol.item_size = 8

        # private members
        symbol._client = client
        symbol._offset = 0
        symbol._file_address = None

        # getting more data out from lldb
        if lldb_address is None:
            lldb_address = client.target.ResolveLoadAddress(int(symbol) & 0xFFFFFFFFFFFFFFFF)
        if lldb_type is None:
            lldb_type = lldb_address.symbol.type
        symbol.type_ = lldb_type
        symbol.lldb_address = lldb_address
        symbol.lldb_symbol = lldb_symbol

        for method_name in Symbol.PROXY_METHODS:
            getattr(symbol.__class__, method_name).__doc__ = \
                getattr(client, method_name).__doc__

        return symbol

    @property
    def id(self) -> HildaSymbolId:
        return (self.lldb_name, int(self))

    @property
    def lldb_name(self) -> Optional[str]:
        return self.lldb_symbol.GetName() if self.lldb_symbol is not None else None

    @cached_property
    def file_address(self) -> int:
        """
        Get symbol file address (address without ASLR)
        :return: File address
        """
        return self.lldb_address.file_addr

    @cached_property
    def filename(self):
        return self.lldb_address.module.file.basename

    @property
    def objc_class(self) -> Class:
        """
        Get the objc class of the respected symbol
        :return: Class
        """
        return Class(self._client, self.objc_call('class'))

    @property
    def objc_symbol(self):
        """
        Get an ObjectiveC symbol of the same address
        :return: Object representing the ObjectiveC symbol
        """
        return self._client.objc_symbol(self)

    @property
    def cf_description(self) -> str:
        """
        Get output from CFCopyDescription()
        :return: CFCopyDescription()'s output as a string
        """
        return self._client.symbols.CFCopyDescription(self).po()

    @property
    def name(self) -> str:
        symbol_info = int(self._client.po(f'[{self._client._object_identifier} symbolForAddress:{self}]', '__int128'))
        arg1 = symbol_info & 0xffffffffffffffff
        arg2 = symbol_info >> 64
        return self._client.symbols.CSSymbolGetName(arg1, arg2).peek_str()

    @contextmanager
    def change_item_size(self, new_item_size: int) -> None:
        """
        Temporarily change item size
        :param new_item_size: Temporary item size
        """
        save_item_size = self.item_size
        self.item_size = new_item_size
        try:
            yield
        finally:
            self.item_size = save_item_size

    def py(self) -> CfSerializable:
        return self._client.decode_cf(self)

    def peek(self, count: int):
        return self._client.peek(self, count)

    def poke(self, buf: bytes) -> None:
        return self._client.poke(self, buf)

    def poke_text(self, code: str) -> int:
        return self._client.poke_text(self, code)

    def peek_str(self) -> str:
        return self._client.peek_str(self)

    def peek_std_str(self) -> str:
        return self._client.peek_std_str(self)

    def monitor(self, **args):
        return self._client.monitor(self, **args)

    def watch(self, **args):
        return self._client.watchpoints.add(self, **args)

    def bp(self, callback=None, **args):
        return self._client.bp(self, callback, **args)

    def disass(self, size, **args) -> lldb.SBInstructionList:
        return self._client.disass(self, self.peek(size), **args)

    def po(self, cast: Optional[str] = None) -> str:
        return self._client.po(self, cast=cast)

    def objc_call(self, selector: str, *params) -> Any:
        return self._client.objc_call(self, selector, *params)

    def close(self) -> None:
        """ Construct compliance. """
        pass

    def seek(self, offset: int, whence: int = os.SEEK_SET) -> None:
        """ Construct compliance. """
        if whence == os.SEEK_CUR:
            self._offset += offset
        elif whence == os.SEEK_SET:
            self._offset = offset - self
        else:
            raise OSError('Unsupported whence')

    def read(self, count: int) -> bytes:
        """ Construct compliance. """
        val = (self + self._offset).peek(count)
        self._offset += count
        return val

    def write(self, buf: bytes) -> int:
        """ Construct compliance. """
        val = (self + self._offset).poke(buf)
        self._offset += len(buf)
        return val

    def tell(self) -> int:
        """ Construct compliance. """
        return self + self._offset

    def __add__(self, other):
        try:
            return self._client.symbol(int(self) + other)
        except TypeError:
            return int(self) + other

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        try:
            return self._client.symbol(int(self) - other)
        except TypeError:
            return int(self) - other

    def __rsub__(self, other):
        try:
            return self._client.symbol(other - int(self))
        except TypeError:
            return other - int(self)

    def __mul__(self, other):
        try:
            return self._client.symbol(int(self) * other)
        except TypeError:
            return int(self) * other

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self._client.symbol(int(self) / other)

    def __floordiv__(self, other):
        return self._client.symbol(int(self) // other)

    def __mod__(self, other):
        return self._client.symbol(int(self) % other)

    def __and__(self, other):
        return self._client.symbol(int(self) & other)

    def __or__(self, other):
        return self._client.symbol(int(self) | other)

    def __xor__(self, other):
        return self._client.symbol(int(self) ^ other)

    def __getitem__(self, item):
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        addr = self + item * self.item_size
        return self._client.symbol(
            struct.unpack(self._client.endianness + fmt, self._client.peek(addr, self.item_size))[0])

    def __setitem__(self, item, value):
        fmt = ADDRESS_SIZE_TO_STRUCT_FORMAT[self.item_size]
        value = struct.pack(self._client.endianness + fmt, int(value))
        self._client.poke(self + item * self.item_size, value)

    def __repr__(self):
        address = int(self)
        name = self.lldb_name
        if name is not None:
            return f'Symbol({name}, 0x{address:016X})'
        else:
            return f'AnonymousSymbol(0x{address:016X})'

    def __str__(self):
        return hex(self)

    def __call__(self, *args, **kwargs):
        return self._client.call(self, args)



import ast
import builtins

from IPython.terminal.interactiveshell import TerminalInteractiveShell

from hilda.exceptions import EvaluatingExpressionError
from hilda.hilda_client import HildaClient
from hilda.lldb_importer import lldb


class HIEvents:
    def __init__(self, ip: TerminalInteractiveShell):
        self.shell = ip
        self.hilda_client: HildaClient = self.shell.user_ns['p']

    def pre_run_cell(self, info):
        """
        Enable lazy loading for symbols
        :param info: IPython's CellInfo object
        """
        if info.raw_cell[0] in ['!', '%'] or info.raw_cell.endswith('?'):
            return
        for node in ast.walk(ast.parse(info.raw_cell)):
            if not isinstance(node, ast.Name):
                # we are only interested in names
                continue

            if node.id in locals() or node.id in self.hilda_client.globals or node.id in dir(builtins):
                # That are undefined
                continue

            symbol = self.hilda_client.symbols.get(node.id)
            if symbol is None:
                continue

            try:
                self.hilda_client._add_global(
                    node.id,
                    symbol if symbol.type_ != lldb.eSymbolTypeObjCMetaClass else self.hilda_client.objc_get_class(
                        node.id)
                )
            except EvaluatingExpressionError:
                self.hilda_client.log_warning(
                    f'Process is running. Pause execution in order to resolve "{node.id}"')


def load_ipython_extension(ip: TerminalInteractiveShell):
    hie = HIEvents(ip)
    ip.events.register('pre_run_cell', hie.pre_run_cell)



import logging

SYSLOG_LISTENER_CODE = '''
socat UNIX-RECVFROM:/var/run/syslog, reuseaddr, fork, mode-0777 STDOUT
'''


def open_syslog_socket():
    logging.info(f'please execute the following code on the device for log listening:\n{SYSLOG_LISTENER_CODE}')
    input('> Hit return to resume')



__all__ = ['HildaException', 'SymbolAbsentError', 'EvaluatingExpressionError', 'CreatingObjectiveCSymbolError',
           'ConvertingToNsObjectError', 'ConvertingFromNSObjectError', 'DisableJetsamMemoryChecksError',
           'GettingObjectiveCClassError', 'AccessingRegisterError', 'AccessingMemoryError',
           'AddingLldbSymbolError', 'LLDBError', 'InvalidThreadIndexError']


class HildaException(Exception):
    """ A domain exception for hilda errors. """
    pass


class LLDBError(HildaException):
    """ Wrapper for RAW LLDB errors """
    pass


class SymbolAbsentError(HildaException):
    """ Raise when trying to get a symbol that doesn't exist. """
    pass


class EvaluatingExpressionError(HildaException):
    """ Raise when failing to evaluate an expression. """
    pass


class CreatingObjectiveCSymbolError(HildaException):
    """ Raise when failing to create an ObjectiveC Symbol. """
    pass


class ConvertingToNsObjectError(HildaException):
    """ Raise when failing to convert python object to NS object. """
    pass


class ConvertingFromNSObjectError(HildaException):
    """ Raise when failing to convert NS object to python object. """
    pass


class DisableJetsamMemoryChecksError(HildaException):
    """ Raise when failing to disable jetsam memory checks. """
    pass


class GettingObjectiveCClassError(HildaException):
    """ Raise when failing to get an ObjectiveC class. """
    pass


class AccessingRegisterError(HildaException):
    """ Raise when failing to access a register. """
    pass


class AccessingMemoryError(HildaException):
    """ Raise when failing to access memory. """
    pass


class AddingLldbSymbolError(HildaException):
    """ Raise when failing to convert a LLDB symbol to Hilda's symbol. """
    pass


class InvalidThreadIndexError(HildaException):
    """ Raise when thread idx invalid """
    pass



from datetime import datetime
from typing import Any, Union

import inquirer3
from inquirer3.themes import GreenPassion

CfSerializable = Union[
    dict[str, Any], list, tuple[Any, ...], str, bool, float, bytes, datetime, None]


def selection_prompt(options_list: list):
    question = [inquirer3.List('choice', message='choose device', choices=options_list, carousel=True)]
    result = inquirer3.prompt(question, theme=GreenPassion(), raise_keyboard_interrupt=True)
    return result['choice']



import logging
import subprocess
import sys
from types import ModuleType
from typing import Optional

logger = logging.getLogger(__name__)


def get_lldb_python_path() -> str:
    result = subprocess.run(['lldb', '-P'], capture_output=True, text=True, check=True)
    return result.stdout.strip()


def import_lldb() -> Optional[ModuleType]:
    lldb_python_path = get_lldb_python_path()
    if lldb_python_path not in sys.path:
        sys.path.append(lldb_python_path)
    import lldb
    return lldb


lldb = import_lldb()



from typing import Callable, Generator, Optional, Union

from hilda.lldb_importer import lldb


class HildaWatchpoint:
    """
    Hilda's class representing an LLDB watchpoint, with some optional additional properties
    """

    def __init__(self, hilda, lldb_watchpoint: lldb.SBWatchpoint, where: Optional[int] = None) -> None:
        """
        Initialize a watchpoint.

        :param hilda.hilda_client.HildaClient hilda: Hilda client
        """
        self._hilda = hilda
        self._where = where
        self._callback = None

        # Actual watchpoint from LLDB API
        self.lldb_watchpoint = lldb_watchpoint

    @property
    def where(self) -> Optional[int]:
        """
        A value identifying where the watchpoint was set (when it was created).

        It could be either an address (int) or a Hilda symbol object (Symbol, that inherits from int).
        Note that self.address is similar, but self.where is where the watchpoint was set when it was
        created (using Hilda API), and self.address is the actual address. They should have the same
        value, although self.where may be a Hilda Symbol.
        """
        return self._where

    @property
    def id(self) -> int:
        """ A number identifying the watchpoint. """
        return self.lldb_watchpoint.GetID()

    @property
    def callback(self) -> Optional[Callable]:
        """
        A callback that will be executed when the watchpoint is hit.

        Note that unless the callback explicitly continues (by calling `cont()`), the program will not continue.
        The callback will be invoked as callback(hilda, *args), where hilda is the 'HildaClient'.
        """
        return self._callback

    @callback.setter
    def callback(self, callback: Optional[Callable]) -> None:
        self._callback = callback
        # TODO: Figure out a way to add set this callback programmatically
        self._hilda.lldb_handle_command(f'watchpoint command add -F '
                                        f'lldb.hilda_client.watchpoints._dispatch_watchpoint_callback {self.id}')

    @property
    def condition(self) -> Optional[str]:
        """
        An LLDB expression to make this a conditional watchpoint.
        """
        return self.lldb_watchpoint.GetCondition()

    @condition.setter
    def condition(self, condition: Optional[str]) -> None:
        self.lldb_watchpoint.SetCondition(condition)

    @property
    def address(self) -> int:
        """
        Get the address this watchpoint watches (also see self.where).
        """
        return self.lldb_watchpoint.GetWatchAddress()

    @property
    def size(self) -> int:
        """
        Get the size this watchpoint watches.
        """
        return self.lldb_watchpoint.GetWatchSize()

    @property
    def enabled(self) -> bool:
        """
        Configures whether this watchpoint is enabled or not.
        """
        return self.lldb_watchpoint.IsEnabled()

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self.lldb_watchpoint.SetEnabled(value)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} LLDB:{self.lldb_watchpoint} CALLBACK:{self.callback}>'

    def __str__(self) -> str:
        emoji = '🚨' if self.enabled else '🔕'
        enabled_str = 'enabled' if self.enabled else 'disabled'
        result = f'{emoji} Watchpoint #{self.id} ({enabled_str})\n'

        if self.where is not None:
            result += f'\tWhere: {self.where}\n'

        return result.strip('\n')

    def remove(self) -> None:
        """
        Remove the watchpoint.
        """
        self._hilda.watchpoints.remove(self)


class WatchpointList:
    """
    Manager for `HildaWatchpoint` objects, each one wrapping another native LLDB watchpoint.
    """

    def __init__(self, hilda) -> None:
        """
        Initialize a watchpoint list.

        :param hilda.hilda_client.HildaClient hilda: Hilda client
        """
        self._hilda = hilda
        self._watchpoints = {}

    def __contains__(self, id_or_wp: Union[int, HildaWatchpoint]) -> bool:
        return self.get(id_or_wp) is not None

    def __iter__(self) -> Generator[HildaWatchpoint, None, None]:
        for wp in self._hilda.target.watchpoint_iter():
            yield self[wp.GetID()]

    def __len__(self) -> int:
        return self._hilda.target.GetNumWatchpoints()

    def __getitem__(self, id_or_wp: Union[int, HildaWatchpoint]) -> HildaWatchpoint:
        """
        Get a watchpoint by ID (or the watchpoint itself, though it usually makes little sense)

        :param id_or_wp: Watchpoint's ID (or the watchpoint itself)
        """
        wp = self.get(id_or_wp)
        if wp is None:
            raise KeyError(id_or_wp)

        return wp

    def __delitem__(self, id_or_wp: Union[int, HildaWatchpoint]):
        """
        Remove a watchpoint.

        :param id_or_wp: Watchpoint's ID (or the watchpoint itself)
        """
        self.remove(id_or_wp)

    def __repr__(self) -> str:
        return repr(dict(self.items()))

    def __str__(self) -> str:
        return repr(self)

    def get(self, id_or_wp: Union[int, HildaWatchpoint]) -> Optional[HildaWatchpoint]:
        """
        Get a watchpoint by ID or the watchpoint itself.

        :param id_or_wp: Watchpoint's ID (or the watchpoint itself)
        :return: `HildaWatchpoint` if one exists, or `None` otherwise
        """

        if isinstance(id_or_wp, int):
            wp = self._hilda.target.FindWatchpointByID(id_or_wp)
        elif isinstance(id_or_wp, HildaWatchpoint):
            wp = id_or_wp.lldb_watchpoint
        else:
            raise KeyError(f'Watchpoint "{id_or_wp}" could not be found')

        if not wp.IsValid():
            return None

        wp_id = wp.GetID()
        if wp_id not in self._watchpoints:
            self._hilda.log_debug(f'Found a watchpoint added outside of the Hilda API {wp}')
            self._watchpoints[wp_id] = HildaWatchpoint(self._hilda, wp)

        return self._watchpoints[wp_id]

    def add(self, where: int, size: int = 8, read: bool = True, write: bool = True,
            callback: Optional[Callable] = None, condition: str = None) -> HildaWatchpoint:
        """
        Add a watchpoint.

        :param where: The address of the watchpoint.
        :param size: The size of the watchpoint (the address span to watch).
        :param read: The watchpoint should monitor reads from memory in the specified address (and size).
        :param write: The watchpoint should monitor writes to memory in the specified address (and size).
        :param callback: A callback that will be executed when the watchpoint is hit.
            Note that unless the callback explicitly continues (by calling cont()), the program will not continue.
            The callback will be invoked as callback(hilda, *args), where hilda is the 'HildaClient'.
        :param condition: An LLDB expression to make this a conditional watchpoint.
        :return: The new watchpoint
        """

        error = lldb.SBError()
        wp = self._hilda.target.WatchAddress(where, size, read, write, error)
        if not wp.IsValid():
            raise Exception(f'Failed to create watchpoint at {where} ({error})')

        wp = HildaWatchpoint(self._hilda, wp, where)
        wp.callback = callback
        wp.condition = condition

        self._watchpoints[wp.id] = wp

        self._hilda.log_info(f'Watchpoint #{wp.id} has been set')
        return wp

    def remove(self, id_or_wp: Union[int, HildaWatchpoint]) -> None:
        """
        Remove a watchpoint.

        :param id_or_wp: Watchpoint's ID (or the watchpoint itself)
        """
        wp = self[id_or_wp]
        watchpoint_id = wp.id
        self._hilda.target.DeleteWatchpoint(watchpoint_id)
        self._hilda.log_debug(f'Watchpoint #{watchpoint_id} has been removed')

    def clear(self) -> None:
        """
        Remove all watchpoints
        """
        for wp in list(self):
            self.remove(wp)

    def show(self) -> None:
        """ Show existing watchpoints. """
        if len(self) == 0:
            self._hilda.log_info('No watchpoints')
        for wp in self:
            self._hilda.log_info(wp)

    def items(self) -> Generator[tuple[int, HildaWatchpoint], None, None]:
        """
        Get a watchpoint ID and watchpoint object tuple for every watchpoint.
        """
        return ((wp.id, wp) for wp in self)

    def keys(self) -> Generator[int, None, None]:
        """
        Get the watchpoint ID for every watchpoint.
        """
        return (wp.id for wp in self)

    def values(self) -> Generator[HildaWatchpoint, None, None]:
        """
        Get the watchpoint object for every watchpoint.
        """
        return (wp for wp in self)

    def _dispatch_watchpoint_callback(self, frame, wp, internal_dict) -> None:
        """
        Route the watchpoint callback the specific watchpoint callback.
        """
        watchpoint_id = wp.GetID()
        self._hilda._bp_frame = frame
        try:
            callback = self[watchpoint_id].callback
            if callback is not None:
                callback(self._hilda, frame, None, self[watchpoint_id])
        finally:
            self._hilda._bp_frame = None



import pytest

from hilda.exceptions import GettingObjectiveCClassError


def test_get_objc_class_error(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    with pytest.raises(GettingObjectiveCClassError):
        hilda_client.objc_get_class('classthatdoesntexist')


def test_call_symbol_with_bytes(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    b = bytes(range(0x100))
    with hilda_client.safe_malloc(len(b)) as a:
        hilda_client.symbols.memcpy(a, b, len(b))
        assert a.peek(len(b)) == b


def test_lsof(hilda_client):
    """
    :param hilda.hilda_client.HildaClient hilda_client: Hilda client.
    """
    temp_dir_url = hilda_client.objc_get_class('NSFileManager').defaultManager().objc_call('temporaryDirectory')
    temp_url = temp_dir_url.objc_call('URLByAppendingPathComponent:', hilda_client.ns('temp.txt')).objc_symbol
    c_file_path = temp_url.path.objc_call('cString')
    file_path = c_file_path.peek_str()
    file_handle = hilda_client.symbols.mkstemp(c_file_path)
    max_path_len = hilda_client.evaluate_expression('PATH_MAX')
    try:
        with hilda_client.safe_malloc(max_path_len) as real_path:
            hilda_client.symbols.realpath(file_path, real_path)
            assert hilda_client.lsof()[file_handle] == real_path.peek_str()
    finally:
        hilda_client.symbols.close(file_handle)
        hilda_client.symbols.unlink(file_path)



from construct import Array, BitsInteger, BitStruct, Bytes, Enum, Flag, Hex, Int32ul, Int64ul, Padding, Struct, Tell

from hilda.lldb_importer import lldb
from hilda.symbol import SymbolFormatField

malloc_zone_t = Struct(
    'RESERVED_1_CFALLOCATOR' / Hex(Int64ul),
    'RESERVED_2_CFALLOCATOR' / Hex(Int64ul),
    'size' / SymbolFormatField(lldb.hilda_client),
    'malloc' / SymbolFormatField(lldb.hilda_client),
    'calloc' / SymbolFormatField(lldb.hilda_client),
    'valloc' / SymbolFormatField(lldb.hilda_client),
    'free' / SymbolFormatField(lldb.hilda_client),
    'realloc' / SymbolFormatField(lldb.hilda_client),
    'destroy' / SymbolFormatField(lldb.hilda_client),
    'zone_name' / SymbolFormatField(lldb.hilda_client),
    'batch_malloc' / SymbolFormatField(lldb.hilda_client),
    'batch_free' / SymbolFormatField(lldb.hilda_client),
    'introspect' / SymbolFormatField(lldb.hilda_client),
    'version' / SymbolFormatField(lldb.hilda_client),
    'memalign' / SymbolFormatField(lldb.hilda_client),
    'free_definite_size' / SymbolFormatField(lldb.hilda_client),
    'pressure_relief' / SymbolFormatField(lldb.hilda_client),
    'claimed_address' / SymbolFormatField(lldb.hilda_client),
    'try_free_default' / SymbolFormatField(lldb.hilda_client),
    'malloc_with_options' / SymbolFormatField(lldb.hilda_client),
    'malloc_type_malloc' / SymbolFormatField(lldb.hilda_client),
    'malloc_type_calloc' / SymbolFormatField(lldb.hilda_client),
    'malloc_type_realloc' / SymbolFormatField(lldb.hilda_client),
    'malloc_type_memalign' / SymbolFormatField(lldb.hilda_client),
    'malloc_type_malloc_with_options' / SymbolFormatField(lldb.hilda_client),
)

nanov2_statistics_t = Struct(
    'allocated_regions' / Hex(Int32ul),
    'region_addresses_clhashes' / Hex(Int32ul),
    # 'size_class_statistics' / Array(16, nanov2_size_class_statistics)
)

nanozonev2_s = Struct(
    'address' / Tell,
    'basic_zone' / malloc_zone_t,
    # pad used to mprotect the first page
    Padding(0x4000 - malloc_zone_t.sizeof()+32),
    # metadata of arena blocks
    'current_block' / Hex(Int64ul),
    # current_block_lock omitted
    Padding(0x2fd4),
    'delegate_allocations' / Hex(Int32ul),
    'debug_flags' / Hex(Int64ul),
    'aslr_cookie' / Hex(Int64ul),
    'aslr_cookie_aligned' / Hex(Int64ul),
    'slot_freelist_cookies' / Hex(Int64ul),
    'helper_zone' / Hex(Int64ul),
    'block_lock' / Hex(Int32ul),
    'regions_lock' / Hex(Int32ul),
    'first_region_base_ptr' / Hex(Int64ul),
    'current_region_next_arena' / Hex(Int64ul),
    'madvise_lock' / Hex(Int64ul),
    'nanov2_statistics_t' / nanov2_statistics_t
)

nanov2_size_class_statistics = Struct(
    'total_allocations' / Hex(Int64ul),
    'total_frees' / Hex(Int64ul),
    'madvised_blocks' / Hex(Int64ul),
    'madvise_races' / Hex(Int64ul),
)


class NanoV2Zone:

    def __init__(self, nanov2_base_addr, helper_zone_ptr):
        self.nanov2_base_addr = nanov2_base_addr
        self.nanov2_struct = nanozonev2_s.parse_stream(self.nanov2_base_addr)
        if not self._sanity_check(helper_zone_ptr):
            self.nanov2_struct = None

    def _sanity_check(self, helper_zone_ptr):
        return helper_zone_ptr == int(str(self.nanov2_struct.helper_zone), 16)

    def get_nanov2_struct(self):
        return self.nanov2_struct


class NanoV2Arena:
    # You would have to iterate for 64 times to dump everything. But not going to dump 64MB...
    # Found at nanov2_malloc.c
    size_per_slot = [16, 32, 48, 64, 80, 96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 256]
    blocks_per_size = [2, 10, 11, 10, 5, 3, 3, 4, 3, 2, 2, 2, 2, 2, 1, 2]
    slots_per_size = [1024, 512, 341, 256, 204, 170, 146, 128, 113, 102, 93, 85, 78, 73, 68, 64]

    def __init__(self, arena_base_addr):
        self.nanov2_arena_struct = Struct(
            # Note that this only parses 1 CPU memblock (1MB). Modify it to be 64.
            *[
                f"blocks_{idx}" / Array(self.blocks_per_size[idx],
                                        Struct('content' /
                                               Array(self.slots_per_size[idx],
                                                     Struct('Q' / Bytes(self.size_per_slot[idx])
                                                            )
                                                     )
                                               ))
                for idx in range(len(self.blocks_per_size))
            ]
        )
        self.arena_struct = self.nanov2_arena_struct.parse_stream(arena_base_addr)

    def get_arena_struct(self):
        return self.arena_struct


class NanoV2ArenaMetadataBlock:
    # This is used by libmalloc to store the metadata of the arena blocks which is used during allocations.
    # It is a matrix because we have a list of block sizes * CPU_NUMBER.

    NEXT_SLOT = Enum(BitsInteger(11),
                     SLOT_NULL=0,
                     SLOT_GUARD=0x7fa,
                     SLOT_BUMP=0x7fb,
                     SLOT_FULL=0x7fc,
                     SLOT_CAN_MADVISE=0x7fd,
                     SLOT_MADVISING=0x7fe,
                     SLOT_MADVISED=0x7ff
                     )

    block_meta_t = BitStruct(
        'in_use' / Flag,
        'gen_count' / Hex(BitsInteger(10)),
        'free_count' / Hex(BitsInteger(10)),
        'next_slot' / NEXT_SLOT,
    )

    def __init__(self, curr_block_addr):
        self.nanov2_arena_metablock_t = Struct(
            # We can't directly parse the block_meta with the CStruct.
            # It has to be Int32ul first because BitStruct does not handle unalignment properly.
            "arena_block_meta" / Array(4096, Hex(Int32ul))
        )
        self.curr_block_addr = curr_block_addr
        self.arena_metadata_block = self.nanov2_arena_metablock_t.parse_stream(curr_block_addr)

    def dump_arena_metadata_block(self):
        block_idx = 0
        for meta_block in self.arena_metadata_block.arena_block_meta:
            hex_meta_contents = str(meta_block)[2:]
            block_meta_struct = self.block_meta_t.parse(bytes.fromhex(hex_meta_contents))
            if block_meta_struct.in_use:
                print(f"Arena block index {block_idx} (CPU {int(block_idx / 64) - 1}): next: {block_meta_struct.next_slot}")
            block_idx = block_idx + 1


RACK_TYPE = Enum(Int32ul,
                 RACK_TYPE_NONE=0,
                 RACK_TYPE_TINY=1,
                 RACK_TYPE_SMALL=2,
                 RACK_TYPE_MEDIUM=3
                 )

region_hash_generation_t = Struct(
    'num_regions_allocated' / Hex(Int64ul),
    'num_regions_allocated_shift' / Hex(Int64ul),
    'hashed_regions_ptr' / Hex(Int64ul),
    'nextgen_ptr' / Hex(Int64ul),
)

rack_s = Struct(
    # Keep in mind this struct is MALLOC_CACHE_ALIGN(128) aligned.
    'address' / Tell,
    'region_lock' / Hex(Int32ul),
    'type' / RACK_TYPE,
    'num_regions' / Hex(Int64ul),
    'num_regions_dealloc' / Hex(Int64ul),
    'region_generation_ptr' / Hex(Int64ul),
    'rg' / Array(2, region_hash_generation_t),
    'initial_regions' / Array(64, Hex(Int64ul)),
    'num_magazines' / Hex(Int32ul),
    'num_magazines_mask' / Hex(Int32ul),
    'num_magazines_mask_shift' / Hex(Int32ul),
    'debug_flags' / Hex(Int32ul),
    'magazines' / Hex(Int64ul),
    'cookie' / Hex(Int64ul),
    'last_madvise' / Hex(Int64ul),
)

szone_t = Struct(
    'address' / Tell,
    'basic_zone' / malloc_zone_t,
    # it seems like the offset is 0x4068 from the start of the basic_zone to the tiny rack
    Padding(0x4000 - malloc_zone_t.sizeof() + 0x68),
    'cpu_id_key' / Hex(Int64ul),
    'debug_flags' / Hex(Int64ul),
    'log_address' / Hex(Int64ul),
    'tiny_rack' / rack_s,
    Padding(0x78),
    'small_rack' / rack_s,
    Padding(0x78),
    'medium_rack' / rack_s,
    'large_szone_lock' / Hex(Int32ul),
    'num_large_objects_in_use' / Hex(Int32ul),
    'num_large_entries' / Hex(Int32ul),
    'large_entries' / Hex(Int64ul),
    'num_bytes_in_large_objects' / Hex(Int64ul),
    # for macOS, CONFIG_LARGE_CACHE is True. For iOS is False (TODO)
)


class HelperZone:
    def __init__(self, helper_zone_ptr):
        self.szone_struct = szone_t.parse_stream(helper_zone_ptr)

    def get_helper_zone(self):
        return self.szone_struct


def _get_nanov2_zone():
    hilda = lldb.hilda_client
    default_nanozone_ptr = hilda.symbols.malloc_zones[0][0]
    # we need the helper_zone_ptr to perform a sanity check on the parsing.
    helper_zone_ptr = hilda.symbols.malloc_zones[0][1]
    return NanoV2Zone(default_nanozone_ptr, helper_zone_ptr)


def dump_helper_zone():
    hilda = lldb.hilda_client
    helper_zone_ptr = hilda.symbols.malloc_zones[0][1]
    helper_zone = HelperZone(helper_zone_ptr)
    print(helper_zone.get_helper_zone())


def dump_nanov2_zone():
    nano_zone_class = _get_nanov2_zone()
    print(nano_zone_class.get_nanov2_struct())


def dump_used_arena():
    """
    Arena first block is the metadata
    Afterwards, arena consists of 64 logical blocks (1 per each CPU) that each of them is 1MB.
    Each CPU mem block is -> blocks_per_size = [2,10,11,10,5,3,3,4,3,2,2,2,2,2,1,2] -> 64 blocks
    Something like this:

    (remember, 1 block is 16k)
        ----------------------------------
        ARENA_METADATA_BLOCK - 16K
        ----------------------------------
                    (it is inline, this arrow is not a ptr. just drawing purposes.)
        CPU_0_MEM_BLOCK - 1MB -----------> | BLOCK_16 * 2
                                           BLOCK_32 * 10
                                           BLOCK_48 * 11
                                           .... (check blocks_per_size)
        ----------------------------------
        CPU_1_MEM_BLOCK - 1MB
        ----------------------------------
        CPU_X_MEM_BLOCK - 1MB (There are up to 64!)
        ----------------------------------

        Which makes an arena 64CPU mem blocks * 1 mb = 64mb + arena metadata block = 65mb.
        A region consists of 8 arenas which results in 64mb * 8 = 512mb (no metadata accounted)
    """

    # Note there can be up to NANOV2_ARENAS_PER_REGION arenas (8). We will only dump 1 of them.
    hilda = lldb.hilda_client
    nano_zone_class = _get_nanov2_zone()
    # arena base addr is always 0x0000600000000000 + aslr_cookie_enabled + 0x4000 (first block is the arena metablock)
    first_region_base_ptr_int = int(str(nano_zone_class.get_nanov2_struct().first_region_base_ptr), 16)
    aslr_cookie_aligned_int = int(str(nano_zone_class.get_nanov2_struct().aslr_cookie_aligned), 16)
    arena_addr = first_region_base_ptr_int + aslr_cookie_aligned_int + 0x4000
    nanov2_arena_struct = NanoV2Arena(hilda.symbol(arena_addr))
    with open('/tmp/arena_contents.txt', 'w') as arena_dump_file:
        print("Dumping 1MB of arena (CPU_0 mem block) contents into /tmp/arena_contents.txt.")
        arena_dump_file.write(str(nanov2_arena_struct.get_arena_struct()))


def dump_nanov2_block_metadata():
    # The metadata block is the first logical block in the arena. Arena metadata is at arena_addr + aslr_cookie_aligned
    hilda = lldb.hilda_client
    nano_zone_class = _get_nanov2_zone()
    first_region_base_ptr_int = int(str(nano_zone_class.get_nanov2_struct().first_region_base_ptr), 16)
    aslr_cookie_aligned_int = int(str(nano_zone_class.get_nanov2_struct().aslr_cookie_aligned), 16)
    metadata_addr = first_region_base_ptr_int + aslr_cookie_aligned_int
    print(f"Arena metadata can be found at {metadata_addr:x}")
    current_block_symbol = hilda.symbol(metadata_addr)
    meta_block_class = NanoV2ArenaMetadataBlock(current_block_symbol)
    meta_block_class.dump_arena_metadata_block()



import logging
import os
import sys
from abc import ABC, abstractmethod
from threading import Thread
from typing import Optional

from hilda.exceptions import LLDBError
from hilda.hilda_client import HildaClient
from hilda.lldb_importer import lldb

TIMEOUT = 1
lldb.hilda_client = None

logger = logging.getLogger(__name__)


def execute(cmd: str) -> int:
    logging.debug(f'executing: {cmd}')
    return os.system(cmd)


class LLDBListenerThread(Thread, ABC):

    def __init__(self):
        super().__init__()
        lldb.SBDebugger.Initialize()
        self.debugger: lldb.SBDebugger = lldb.SBDebugger.Create()
        self.listener: lldb.SBListener = self.debugger.GetListener()
        self.error: lldb.SBError = lldb.SBError()
        self.debugger.SetAsync(True)
        self.target: lldb.SBTarget = self._create_target()
        self.process: lldb.SBProcess = self._create_process()
        self._check_success()
        self.should_quit = False

    @abstractmethod
    def _create_target(self) -> lldb.SBTarget:
        pass

    @abstractmethod
    def _create_process(self) -> lldb.SBProcess:
        pass

    def _check_success(self) -> None:
        if self.error.Success():
            return
        raise LLDBError(self.error.description)

    def _process_stdout(self) -> None:
        stdout = self.process.GetSTDOUT(1024)
        while stdout:
            if lldb.hilda_client is not None and lldb.hilda_client.configs.enable_stdout_stderr:
                sys.stdout.write(stdout)
            stdout = self.process.GetSTDOUT(1024)

    def _process_stderr(self) -> None:
        stderr = self.process.GetSTDERR(1024)
        while stderr:
            if lldb.hilda_client is not None and lldb.hilda_client.configs.enable_stdout_stderr:
                sys.stderr.write(stderr)
            stderr = self.process.GetSTDERR(1024)

    def _get_stopped_threads(self, reason: Optional[int] = None) -> list[lldb.SBThread]:
        if reason is None:
            stop_reasons = [
                lldb.eStopReasonSignal, lldb.eStopReasonException,
                lldb.eStopReasonBreakpoint, lldb.eStopReasonWatchpoint,
                lldb.eStopReasonPlanComplete, lldb.eStopReasonTrace,
            ]
        else:
            stop_reasons = [reason]

        return [thread for thread in self.process if thread.GetStopReason() in stop_reasons]

    def _set_selected_thread_to_stopped_thread(self) -> None:
        stopped_threads = self._get_stopped_threads()
        if len(stopped_threads) < 1:
            return
        thread = stopped_threads[0]
        self.process.SetSelectedThread(thread)

    def run(self):
        event = lldb.SBEvent()
        last_state = lldb.eStateStopped
        while not self.should_quit:
            if not self.listener.WaitForEvent(TIMEOUT, event):
                continue
            if not lldb.SBProcess.EventIsProcessEvent(event):
                continue

            event_type = event.GetType()
            if event_type & lldb.SBProcess.eBroadcastBitSTDOUT:
                self._process_stdout()
            if event_type & lldb.SBProcess.eBroadcastBitSTDERR:
                self._process_stderr()

            state = self.process.GetStateFromEvent(event)
            if state == lldb.eStateDetached:
                logger.debug('Process Detached')
                self.should_quit = True
            elif state == lldb.eStateExited:
                logger.info(f'Process Exited with status {self.process.GetExitStatus()}')
                self.should_quit = True
            elif state == lldb.eStateRunning and last_state == lldb.eStateStopped:
                logger.debug('Process Continued')
            elif state == lldb.eStateStopped and last_state == lldb.eStateRunning:
                logger.debug('Process Stopped')
                self._set_selected_thread_to_stopped_thread()

            last_state = state


class LLDBRemote(LLDBListenerThread):
    def __init__(self, address: str, port: int = 1234):
        self.url_connect = f'connect://{address}:{port}'
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTarget('')

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Connecting to "{self.url_connect}"')
        return self.target.ConnectRemote(self.listener, self.url_connect, None, self.error)


class LLDBAttachPid(LLDBListenerThread):

    def __init__(self, pid: int):
        self.pid = pid
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(None, None)

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Attaching to {self.pid}')
        return self.target.AttachToProcessWithID(self.listener, self.pid, self.error)


class LLDBAttachName(LLDBListenerThread):

    def __init__(self, proc_name: str, wait_for: bool = False):
        self.proc_name = proc_name
        self.wait_for = wait_for
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(None, None)

    def _create_process(self) -> lldb.SBProcess:
        logger.debug(f'Attaching to {self.proc_name}')
        return self.target.AttachToProcessWithName(self.listener, self.proc_name, self.wait_for, self.error)


class LLDBLaunch(LLDBListenerThread):
    def __init__(self, exec_path: str, argv: Optional[list[str]] = None, envp: Optional[list[str]] = None,
                 stdin: Optional[str] = None,
                 stdout: Optional[str] = None, stderr: Optional[str] = None, wd: Optional[str] = None,
                 flags: Optional[int] = 0):
        self.exec_path = exec_path
        self.stdout = stdout
        self.stdin = stdin
        self.stderr = stderr
        self.flags = flags
        self.argv = argv
        self.envp = envp
        self.working_directory = wd
        super().__init__()

    def _create_target(self) -> lldb.SBTarget:
        return self.debugger.CreateTargetWithFileAndArch(self.exec_path, lldb.LLDB_ARCH_DEFAULT)

    def _create_process(self) -> lldb.SBProcess:
        # Launch(SBTarget self, SBListener listener, char const ** argv, char const ** envp,
        # char const * stdin_path, char const * stdout_path, char const * stderr_path, char const * working_directory,
        # uint32_t launch_flags, bool stop_at_entry, SBError error) -> SBProcess
        logger.debug(f'Launching process  {self.exec_path}')
        return self.target.Launch(self.listener, self.argv, self.envp,
                                  self.stdin, self.stdout, self.stderr, self.working_directory,
                                  self.flags, True,
                                  self.error)


def _get_hilda_client_from_sbdebugger(debugger: lldb.SBDebugger) -> HildaClient:
    hilda_client = HildaClient(debugger)
    lldb.hilda_client = hilda_client
    hilda_client.init_dynamic_environment()
    return hilda_client


def create_hilda_client_using_remote_attach(
        hostname: str, port: int) -> HildaClient:
    lldb_t = LLDBRemote(hostname, port)
    lldb_t.start()
    return _get_hilda_client_from_sbdebugger(lldb_t.debugger)


def create_hilda_client_using_launch(
        exec_path: str, argv: Optional[list] = None, envp: Optional[list] = None, stdin: Optional[str] = None,
        stdout: Optional[str] = None, stderr: Optional[str] = None, wd: Optional[str] = None,
        flags: Optional[int] = 0) -> HildaClient:
    lldb_t = LLDBLaunch(exec_path, argv, envp, stdin, stdout, stderr, wd, flags)
    lldb_t.start()
    return _get_hilda_client_from_sbdebugger(lldb_t.debugger)


def create_hilda_client_using_attach_by_pid(pid: Optional[int] = None) -> HildaClient:
    lldb_t = LLDBAttachPid(pid)
    lldb_t.start()
    return _get_hilda_client_from_sbdebugger(lldb_t.debugger)


def create_hilda_client_using_attach_by_name(name: Optional[str] = None, wait_for: bool = False) -> HildaClient:
    lldb_t = LLDBAttachName(name, wait_for)
    lldb_t.start()
    return _get_hilda_client_from_sbdebugger(lldb_t.debugger)



from construct import Array, Int8ul, Int32ul, Int64ul, Struct

uuid_t = Struct(
    'time_low' / Int64ul,
    'time_mid' / Int32ul,
    'time_hi_and_version' / Int32ul,
    'clock_seq_hi_and_reserved' / Int8ul,
    'clock_seq_low' / Int8ul,
    'node' / Array(6, Int8ul)
)
